<?php
  if (!extension_loaded('soap')) die('skip soap extension not available');
  /* xml parser is required by test_schema.inc */
  if (!extension_loaded('xml'))  die('skip xml extension not available');
?>
