/* This is a template RC file.
 * Do not edit with MSVC */
#ifdef APSTUDIO_INVOKED
# error Do not edit with MSVC
#endif

#include "winresrc.h"
#include "main/php_version.h"

LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

#ifndef THANKS_GUYS
# define THANKS_GUYS ""
#endif

#ifdef WANT_LOGO
0 ICON win32\build\php.ico
#endif

#ifndef INTERNAL_NAME /* e.g. 'PHAR extension', 'CGI SAPI' */
# ifdef FILE_DESCRIPTION
#define INTERNAL_NAME FILE_DESCRIPTION /* e.g. 'PHP Script Interpreter', 'GD imaging' */
# else
#define INTERNAL_NAME FILE_NAME /* e.g. 'php7ts.dll', 'php_bz2.dll' */
# endif
#endif

#ifndef URL
#define URL "http://www.php.net/"
#endif

#ifndef EXT_VERSION
#define EXT_VERSION PHP_VERSION
#endif

#ifndef EXT_FILE_VERSION
#define EXT_FILE_VERSION PHP_MAJOR_VERSION,PHP_MINOR_VERSION,PHP_RELEASE_VERSION
#endif

VS_VERSION_INFO VERSIONINFO
 FILEVERSION EXT_FILE_VERSION
 PRODUCTVERSION PHP_MAJOR_VERSION,PHP_MINOR_VERSION,PHP_RELEASE_VERSION
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS (VS_FF_DEBUG|VS_FF_SPECIALBUILD)
#else
 FILEFLAGS 0x0L
#endif
 FILEOS VOS__WINDOWS32
 FILETYPE VFT_DLL
 FILESUBTYPE VFT2_UNKNOWN
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", THANKS_GUYS
            VALUE "CompanyName", "The PHP Group"
#ifdef _DEBUG
            VALUE "FileDescription", FILE_DESCRIPTION " (DEBUG)"
#else
            VALUE "FileDescription", FILE_DESCRIPTION
#endif
            VALUE "FileVersion", EXT_VERSION
            VALUE "InternalName", INTERNAL_NAME
            VALUE "LegalCopyright", "Copyright © The PHP Group"
            VALUE "LegalTrademarks", "PHP"
            VALUE "OriginalFilename", FILE_NAME
            VALUE "ProductName", "PHP"
            VALUE "ProductVersion", PHP_VERSION
#ifdef _DEBUG
            VALUE "SpecialBuild", "Debug build"
#endif
            VALUE "URL", URL
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#ifdef MC_INCLUDE
#include MC_INCLUDE
#endif
