/*
 +----------------------------------------------------------------------+
 | PHP Version 7                                                        |
 +----------------------------------------------------------------------+
 | This source file is subject to version 3.01 of the PHP license,      |
 | that is bundled with this package in the file LICENSE, and is        |
 | available through the world-wide-web at the following url:           |
 | http://www.php.net/license/3_01.txt                                  |
 | If you did not receive a copy of the PHP license and are unable to   |
 | obtain it through the world-wide-web, please send a note to          |
 | license@php.net so we can mail you a copy immediately.               |
 +----------------------------------------------------------------------+
 | Authors: Sara Golemon <pollita@php.net>                              |
 +----------------------------------------------------------------------+
*/

#ifndef PHP_INTL_CONVERTER_H
#define PHP_INTL_CONVERTER_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"

int php_converter_minit(INIT_FUNC_ARGS);

#endif /* PHP_INTL_CONVERTER_H */
