UCHARDIRECTION(LEFT_TO_RIGHT)
UCHARDIRECTION(RIGHT_TO_LEFT)
UCHARDIRECTION(EUROPEAN_NUMBER)
UCHARDIRECTION(EUROPEAN_NUMBER_SEPARATOR)
UCHARDIRECTION(EUROPEAN_NUMBER_TERMINATOR)
UCHARDIRECTION(ARABIC_NUMBER)
UCHARDIRECTION(COMMON_NUMBER_SEPARATOR)
UCHARDIRECTION(BLOCK_SEPARATOR)
UCHARDIRECTION(SEGMENT_SEPARATOR)
UCHARDIRECTION(WHITE_SPACE_NEUTRAL)
UCHARDIRECTION(OTHER_NEUTRAL)
UCHARDIRECTION(LEFT_TO_RIGHT_EMBEDDING)
UCHARDIRECTION(LEFT_TO_RIGHT_OVERRIDE)
UCHARDIRECTION(RIGHT_TO_LEFT_ARABIC)
UCHARDIRECTION(RIGHT_TO_LEFT_EMBEDDING)
UCHARDIRECTION(RIGHT_TO_LEFT_OVERRIDE)
UCHARDIRECTION(POP_DIRECTIONAL_FORMAT)
UCHARDIRECTION(DIR_NON_SPACING_MARK)
UCHARDIRECTION(BOUNDARY_NEUTRAL)
#if U_ICU_VERSION_MAJOR_NUM >= 52
UCHARDIRECTION(FIRST_STRONG_ISOLATE)
UCHARDIRECTION(LEFT_TO_RIGHT_ISOLATE)
UCHARDIRECTION(RIGHT_TO_LEFT_ISOLATE)
UCHARDIRECTION(POP_DIRECTIONAL_ISOLATE)
#endif /* ICU >= 52 */
UCHARDIRECTION(CHAR_DIRECTION_COUNT)
