EXPORTS
php_pcre2_callout_enumerate
php_pcre2_code_copy
php_pcre2_code_copy_with_tables
php_pcre2_code_free
php_pcre2_compile
php_pcre2_compile_context_copy
php_pcre2_compile_context_create
php_pcre2_compile_context_free
php_pcre2_config
php_pcre2_convert_context_copy
php_pcre2_convert_context_create
php_pcre2_convert_context_free
php_pcre2_dfa_match
php_pcre2_general_context_copy
php_pcre2_general_context_create
php_pcre2_general_context_free
php_pcre2_get_error_message
php_pcre2_get_mark
php_pcre2_get_ovector_pointer
php_pcre2_get_ovector_count
php_pcre2_get_startchar
php_pcre2_jit_compile
php_pcre2_jit_match
php_pcre2_jit_free_unused_memory
php_pcre2_jit_stack_assign
php_pcre2_jit_stack_create
php_pcre2_jit_stack_free
php_pcre2_maketables
php_pcre2_match
php_pcre2_match_context_copy
php_pcre2_match_context_create
php_pcre2_match_context_free
php_pcre2_match_data_create
php_pcre2_match_data_create_from_pattern
php_pcre2_match_data_free
php_pcre2_pattern_info
php_pcre2_serialize_decode
php_pcre2_serialize_encode
php_pcre2_serialize_free
php_pcre2_serialize_get_number_of_codes
php_pcre2_set_bsr
php_pcre2_set_callout
php_pcre2_set_character_tables
php_pcre2_set_compile_extra_options
php_pcre2_set_compile_recursion_guard
php_pcre2_set_depth_limit
php_pcre2_set_glob_escape
php_pcre2_set_glob_separator
php_pcre2_set_heap_limit
php_pcre2_set_match_limit
php_pcre2_set_max_pattern_length
php_pcre2_set_newline
php_pcre2_set_parens_nest_limit
php_pcre2_set_offset_limit
php_pcre2_substitute
php_pcre2_substring_copy_byname
php_pcre2_substring_copy_bynumber
php_pcre2_substring_free
php_pcre2_substring_get_byname
php_pcre2_substring_get_bynumber
php_pcre2_substring_length_byname
php_pcre2_substring_length_bynumber
php_pcre2_substring_list_get
php_pcre2_substring_list_free
php_pcre2_substring_nametable_scan
php_pcre2_substring_number_from_name
php_pcre2_set_recursion_limit
php_pcre2_set_recursion_memory_management
