/* Generated from http://icu-project.org/apiref/icu4c/uchar_8h.html#ae40d616419e74ecc7c80a9febab03199 */
UPROPERTY(ALPHABETIC)
UPROPERTY(BINARY_START)
UPROPERTY(ASCII_HEX_DIGIT)
UPROPERTY(BIDI_CONTROL)
UPROPERTY(BIDI_MIRRORED)
UPROPERTY(DASH)
UPROPERTY(DEFAULT_IGNORABLE_CODE_POINT)
UPROPERTY(DEPRECATED)
UPROPERTY(DIACRITIC)
UPROPERTY(EXTENDER)
UPROPERTY(FULL_COMPOSITION_EXCLUSION)
UPROPERTY(GRAPHEME_BASE)
UPROPERTY(GRAPHEME_EXTEND)
UPROPERTY(GRAPHEME_LINK)
UPROPERTY(HEX_DIGIT)
UPROPERTY(HYPHEN)
UPROPERTY(ID_CONTINUE)
UPROPERTY(ID_START)
UPROPERTY(IDEOGRAPHIC)
UPROPERTY(IDS_BINARY_OPERATOR)
UPROPERTY(IDS_TRINARY_OPERATOR)
UPROPERTY(JOIN_CONTROL)
UPROPERTY(LOGICAL_ORDER_EXCEPTION)
UPROPERTY(LOWERCASE)
UPROPERTY(MATH)
UPROPERTY(NONCHARACTER_CODE_POINT)
UPROPERTY(QUOTATION_MARK)
UPROPERTY(RADICAL)
UPROPERTY(SOFT_DOTTED)
UPROPERTY(TERMINAL_PUNCTUATION)
UPROPERTY(UNIFIED_IDEOGRAPH)
UPROPERTY(UPPERCASE)
UPROPERTY(WHITE_SPACE)
UPROPERTY(XID_CONTINUE)
UPROPERTY(XID_START)
UPROPERTY(CASE_SENSITIVE)
UPROPERTY(S_TERM)
UPROPERTY(VARIATION_SELECTOR)
UPROPERTY(NFD_INERT)
UPROPERTY(NFKD_INERT)
UPROPERTY(NFC_INERT)
UPROPERTY(NFKC_INERT)
UPROPERTY(SEGMENT_STARTER)
UPROPERTY(PATTERN_SYNTAX)
UPROPERTY(PATTERN_WHITE_SPACE)
UPROPERTY(POSIX_ALNUM)
UPROPERTY(POSIX_BLANK)
UPROPERTY(POSIX_GRAPH)
UPROPERTY(POSIX_PRINT)
UPROPERTY(POSIX_XDIGIT)
UPROPERTY(CASED)
UPROPERTY(CASE_IGNORABLE)
UPROPERTY(CHANGES_WHEN_LOWERCASED)
UPROPERTY(CHANGES_WHEN_UPPERCASED)
UPROPERTY(CHANGES_WHEN_TITLECASED)
UPROPERTY(CHANGES_WHEN_CASEFOLDED)
UPROPERTY(CHANGES_WHEN_CASEMAPPED)
UPROPERTY(CHANGES_WHEN_NFKC_CASEFOLDED)
UPROPERTY(BINARY_LIMIT)
UPROPERTY(BIDI_CLASS)
UPROPERTY(INT_START)
UPROPERTY(BLOCK)
UPROPERTY(CANONICAL_COMBINING_CLASS)
UPROPERTY(DECOMPOSITION_TYPE)
UPROPERTY(EAST_ASIAN_WIDTH)
UPROPERTY(GENERAL_CATEGORY)
UPROPERTY(JOINING_GROUP)
UPROPERTY(JOINING_TYPE)
UPROPERTY(LINE_BREAK)
UPROPERTY(NUMERIC_TYPE)
UPROPERTY(SCRIPT)
UPROPERTY(HANGUL_SYLLABLE_TYPE)
UPROPERTY(NFD_QUICK_CHECK)
UPROPERTY(NFKD_QUICK_CHECK)
UPROPERTY(NFC_QUICK_CHECK)
UPROPERTY(NFKC_QUICK_CHECK)
UPROPERTY(LEAD_CANONICAL_COMBINING_CLASS)
UPROPERTY(TRAIL_CANONICAL_COMBINING_CLASS)
UPROPERTY(GRAPHEME_CLUSTER_BREAK)
UPROPERTY(SENTENCE_BREAK)
UPROPERTY(WORD_BREAK)
#if U_ICU_VERSION_MAJOR_NUM >= 52
UPROPERTY(BIDI_PAIRED_BRACKET_TYPE)
#endif /* ICU >= 52 */
UPROPERTY(INT_LIMIT)
UPROPERTY(GENERAL_CATEGORY_MASK)
UPROPERTY(MASK_START)
UPROPERTY(MASK_LIMIT)
UPROPERTY(NUMERIC_VALUE)
UPROPERTY(DOUBLE_START)
UPROPERTY(DOUBLE_LIMIT)
UPROPERTY(AGE)
UPROPERTY(STRING_START)
UPROPERTY(BIDI_MIRRORING_GLYPH)
UPROPERTY(CASE_FOLDING)
UPROPERTY(ISO_COMMENT)
UPROPERTY(LOWERCASE_MAPPING)
UPROPERTY(NAME)
UPROPERTY(SIMPLE_CASE_FOLDING)
UPROPERTY(SIMPLE_LOWERCASE_MAPPING)
UPROPERTY(SIMPLE_TITLECASE_MAPPING)
UPROPERTY(SIMPLE_UPPERCASE_MAPPING)
UPROPERTY(TITLECASE_MAPPING)
UPROPERTY(UNICODE_1_NAME)
UPROPERTY(UPPERCASE_MAPPING)
#if U_ICU_VERSION_MAJOR_NUM >= 52
UPROPERTY(BIDI_PAIRED_BRACKET)
#endif /* ICU >= 52 */
UPROPERTY(STRING_LIMIT)
UPROPERTY(SCRIPT_EXTENSIONS)
UPROPERTY(OTHER_PROPERTY_START)
UPROPERTY(OTHER_PROPERTY_LIMIT)
UPROPERTY(INVALID_CODE)
