/* UBlockCode http://icu-project.org/apiref/icu4c/uchar_8h.html#a19f1662c9c33c31593612d05777ea901*/
UBLOCKCODE(NO_BLOCK)
UBLOCKCODE(BASIC_LATIN)
UBLOCKCODE(LATIN_1_SUPPLEMENT)
UBLOCKCODE(LATIN_EXTENDED_A)
UBLOCKCODE(LATIN_EXTENDED_B)
UBLOCKCODE(IPA_EXTENSIONS)
UBLOCKCODE(SPACING_MODIFIER_LETTERS)
UBLOCKCODE(COMBINING_DIACRITICAL_MARKS)
UBLOCKCODE(GREEK)
UBLOCKCODE(CYRILLIC)
UBLOCKCODE(ARMENIAN)
UBLOCKCODE(HEBREW)
UBLOCKCODE(ARABIC)
UBLOCKCODE(SYRIAC)
UBLOCKCODE(THAANA)
UBLOCKCODE(DEVANAGARI)
UBLOCKCODE(BENGALI)
UBLOCKCODE(GURMUKHI)
UBLOCKCODE(GUJARATI)
UBLOCKCODE(ORIYA)
UBLOCKCODE(TAMIL)
UBLOCKCODE(TELUGU)
UBLOCKCODE(KANNADA)
UBLOCKCODE(MALAYALAM)
UBLOCKCODE(SINHALA)
UBLOCKCODE(THAI)
UBLOCKCODE(LAO)
UBLOCKCODE(TIBETAN)
UBLOCKCODE(MYANMAR)
UBLOCKCODE(GEORGIAN)
UBLOCKCODE(HANGUL_JAMO)
UBLOCKCODE(ETHIOPIC)
UBLOCKCODE(CHEROKEE)
UBLOCKCODE(UNIFIED_CANADIAN_ABORIGINAL_SYLLABICS)
UBLOCKCODE(OGHAM)
UBLOCKCODE(RUNIC)
UBLOCKCODE(KHMER)
UBLOCKCODE(MONGOLIAN)
UBLOCKCODE(LATIN_EXTENDED_ADDITIONAL)
UBLOCKCODE(GREEK_EXTENDED)
UBLOCKCODE(GENERAL_PUNCTUATION)
UBLOCKCODE(SUPERSCRIPTS_AND_SUBSCRIPTS)
UBLOCKCODE(CURRENCY_SYMBOLS)
UBLOCKCODE(COMBINING_MARKS_FOR_SYMBOLS)
UBLOCKCODE(LETTERLIKE_SYMBOLS)
UBLOCKCODE(NUMBER_FORMS)
UBLOCKCODE(ARROWS)
UBLOCKCODE(MATHEMATICAL_OPERATORS)
UBLOCKCODE(MISCELLANEOUS_TECHNICAL)
UBLOCKCODE(CONTROL_PICTURES)
UBLOCKCODE(OPTICAL_CHARACTER_RECOGNITION)
UBLOCKCODE(ENCLOSED_ALPHANUMERICS)
UBLOCKCODE(BOX_DRAWING)
UBLOCKCODE(BLOCK_ELEMENTS)
UBLOCKCODE(GEOMETRIC_SHAPES)
UBLOCKCODE(MISCELLANEOUS_SYMBOLS)
UBLOCKCODE(DINGBATS)
UBLOCKCODE(BRAILLE_PATTERNS)
UBLOCKCODE(CJK_RADICALS_SUPPLEMENT)
UBLOCKCODE(KANGXI_RADICALS)
UBLOCKCODE(IDEOGRAPHIC_DESCRIPTION_CHARACTERS)
UBLOCKCODE(CJK_SYMBOLS_AND_PUNCTUATION)
UBLOCKCODE(HIRAGANA)
UBLOCKCODE(KATAKANA)
UBLOCKCODE(BOPOMOFO)
UBLOCKCODE(HANGUL_COMPATIBILITY_JAMO)
UBLOCKCODE(KANBUN)
UBLOCKCODE(BOPOMOFO_EXTENDED)
UBLOCKCODE(ENCLOSED_CJK_LETTERS_AND_MONTHS)
UBLOCKCODE(CJK_COMPATIBILITY)
UBLOCKCODE(CJK_UNIFIED_IDEOGRAPHS_EXTENSION_A)
UBLOCKCODE(CJK_UNIFIED_IDEOGRAPHS)
UBLOCKCODE(YI_SYLLABLES)
UBLOCKCODE(YI_RADICALS)
UBLOCKCODE(HANGUL_SYLLABLES)
UBLOCKCODE(HIGH_SURROGATES)
UBLOCKCODE(HIGH_PRIVATE_USE_SURROGATES)
UBLOCKCODE(LOW_SURROGATES)
UBLOCKCODE(PRIVATE_USE_AREA)
UBLOCKCODE(PRIVATE_USE)
UBLOCKCODE(CJK_COMPATIBILITY_IDEOGRAPHS)
UBLOCKCODE(ALPHABETIC_PRESENTATION_FORMS)
UBLOCKCODE(ARABIC_PRESENTATION_FORMS_A)
UBLOCKCODE(COMBINING_HALF_MARKS)
UBLOCKCODE(CJK_COMPATIBILITY_FORMS)
UBLOCKCODE(SMALL_FORM_VARIANTS)
UBLOCKCODE(ARABIC_PRESENTATION_FORMS_B)
UBLOCKCODE(SPECIALS)
UBLOCKCODE(HALFWIDTH_AND_FULLWIDTH_FORMS)
UBLOCKCODE(OLD_ITALIC)
UBLOCKCODE(GOTHIC)
UBLOCKCODE(DESERET)
UBLOCKCODE(BYZANTINE_MUSICAL_SYMBOLS)
UBLOCKCODE(MUSICAL_SYMBOLS)
UBLOCKCODE(MATHEMATICAL_ALPHANUMERIC_SYMBOLS)
UBLOCKCODE(CJK_UNIFIED_IDEOGRAPHS_EXTENSION_B)
UBLOCKCODE(CJK_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT)
UBLOCKCODE(TAGS)
UBLOCKCODE(CYRILLIC_SUPPLEMENT)
UBLOCKCODE(CYRILLIC_SUPPLEMENTARY)
UBLOCKCODE(TAGALOG)
UBLOCKCODE(HANUNOO)
UBLOCKCODE(BUHID)
UBLOCKCODE(TAGBANWA)
UBLOCKCODE(MISCELLANEOUS_MATHEMATICAL_SYMBOLS_A)
UBLOCKCODE(SUPPLEMENTAL_ARROWS_A)
UBLOCKCODE(SUPPLEMENTAL_ARROWS_B)
UBLOCKCODE(MISCELLANEOUS_MATHEMATICAL_SYMBOLS_B)
UBLOCKCODE(SUPPLEMENTAL_MATHEMATICAL_OPERATORS)
UBLOCKCODE(KATAKANA_PHONETIC_EXTENSIONS)
UBLOCKCODE(VARIATION_SELECTORS)
UBLOCKCODE(SUPPLEMENTARY_PRIVATE_USE_AREA_A)
UBLOCKCODE(SUPPLEMENTARY_PRIVATE_USE_AREA_B)
UBLOCKCODE(LIMBU)
UBLOCKCODE(TAI_LE)
UBLOCKCODE(KHMER_SYMBOLS)
UBLOCKCODE(PHONETIC_EXTENSIONS)
UBLOCKCODE(MISCELLANEOUS_SYMBOLS_AND_ARROWS)
UBLOCKCODE(YIJING_HEXAGRAM_SYMBOLS)
UBLOCKCODE(LINEAR_B_SYLLABARY)
UBLOCKCODE(LINEAR_B_IDEOGRAMS)
UBLOCKCODE(AEGEAN_NUMBERS)
UBLOCKCODE(UGARITIC)
UBLOCKCODE(SHAVIAN)
UBLOCKCODE(OSMANYA)
UBLOCKCODE(CYPRIOT_SYLLABARY)
UBLOCKCODE(TAI_XUAN_JING_SYMBOLS)
UBLOCKCODE(VARIATION_SELECTORS_SUPPLEMENT)
UBLOCKCODE(ANCIENT_GREEK_MUSICAL_NOTATION)
UBLOCKCODE(ANCIENT_GREEK_NUMBERS)
UBLOCKCODE(ARABIC_SUPPLEMENT)
UBLOCKCODE(BUGINESE)
UBLOCKCODE(CJK_STROKES)
UBLOCKCODE(COMBINING_DIACRITICAL_MARKS_SUPPLEMENT)
UBLOCKCODE(COPTIC)
UBLOCKCODE(ETHIOPIC_EXTENDED)
UBLOCKCODE(ETHIOPIC_SUPPLEMENT)
UBLOCKCODE(GEORGIAN_SUPPLEMENT)
UBLOCKCODE(GLAGOLITIC)
UBLOCKCODE(KHAROSHTHI)
UBLOCKCODE(MODIFIER_TONE_LETTERS)
UBLOCKCODE(NEW_TAI_LUE)
UBLOCKCODE(OLD_PERSIAN)
UBLOCKCODE(PHONETIC_EXTENSIONS_SUPPLEMENT)
UBLOCKCODE(SUPPLEMENTAL_PUNCTUATION)
UBLOCKCODE(SYLOTI_NAGRI)
UBLOCKCODE(TIFINAGH)
UBLOCKCODE(VERTICAL_FORMS)
UBLOCKCODE(NKO)
UBLOCKCODE(BALINESE)
UBLOCKCODE(LATIN_EXTENDED_C)
UBLOCKCODE(LATIN_EXTENDED_D)
UBLOCKCODE(PHAGS_PA)
UBLOCKCODE(PHOENICIAN)
UBLOCKCODE(CUNEIFORM)
UBLOCKCODE(CUNEIFORM_NUMBERS_AND_PUNCTUATION)
UBLOCKCODE(COUNTING_ROD_NUMERALS)
UBLOCKCODE(SUNDANESE)
UBLOCKCODE(LEPCHA)
UBLOCKCODE(OL_CHIKI)
UBLOCKCODE(CYRILLIC_EXTENDED_A)
UBLOCKCODE(VAI)
UBLOCKCODE(CYRILLIC_EXTENDED_B)
UBLOCKCODE(SAURASHTRA)
UBLOCKCODE(KAYAH_LI)
UBLOCKCODE(REJANG)
UBLOCKCODE(CHAM)
UBLOCKCODE(ANCIENT_SYMBOLS)
UBLOCKCODE(PHAISTOS_DISC)
UBLOCKCODE(LYCIAN)
UBLOCKCODE(CARIAN)
UBLOCKCODE(LYDIAN)
UBLOCKCODE(MAHJONG_TILES)
UBLOCKCODE(DOMINO_TILES)
UBLOCKCODE(SAMARITAN)
UBLOCKCODE(UNIFIED_CANADIAN_ABORIGINAL_SYLLABICS_EXTENDED)
UBLOCKCODE(TAI_THAM)
UBLOCKCODE(VEDIC_EXTENSIONS)
UBLOCKCODE(LISU)
UBLOCKCODE(BAMUM)
UBLOCKCODE(COMMON_INDIC_NUMBER_FORMS)
UBLOCKCODE(DEVANAGARI_EXTENDED)
UBLOCKCODE(HANGUL_JAMO_EXTENDED_A)
UBLOCKCODE(JAVANESE)
UBLOCKCODE(MYANMAR_EXTENDED_A)
UBLOCKCODE(TAI_VIET)
UBLOCKCODE(MEETEI_MAYEK)
UBLOCKCODE(HANGUL_JAMO_EXTENDED_B)
UBLOCKCODE(IMPERIAL_ARAMAIC)
UBLOCKCODE(OLD_SOUTH_ARABIAN)
UBLOCKCODE(AVESTAN)
UBLOCKCODE(INSCRIPTIONAL_PARTHIAN)
UBLOCKCODE(INSCRIPTIONAL_PAHLAVI)
UBLOCKCODE(OLD_TURKIC)
UBLOCKCODE(RUMI_NUMERAL_SYMBOLS)
UBLOCKCODE(KAITHI)
UBLOCKCODE(EGYPTIAN_HIEROGLYPHS)
UBLOCKCODE(ENCLOSED_ALPHANUMERIC_SUPPLEMENT)
UBLOCKCODE(ENCLOSED_IDEOGRAPHIC_SUPPLEMENT)
UBLOCKCODE(CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C)
UBLOCKCODE(MANDAIC)
UBLOCKCODE(BATAK)
UBLOCKCODE(ETHIOPIC_EXTENDED_A)
UBLOCKCODE(BRAHMI)
UBLOCKCODE(BAMUM_SUPPLEMENT)
UBLOCKCODE(KANA_SUPPLEMENT)
UBLOCKCODE(PLAYING_CARDS)
UBLOCKCODE(MISCELLANEOUS_SYMBOLS_AND_PICTOGRAPHS)
UBLOCKCODE(EMOTICONS)
UBLOCKCODE(TRANSPORT_AND_MAP_SYMBOLS)
UBLOCKCODE(ALCHEMICAL_SYMBOLS)
UBLOCKCODE(CJK_UNIFIED_IDEOGRAPHS_EXTENSION_D)
UBLOCKCODE(ARABIC_EXTENDED_A)
UBLOCKCODE(ARABIC_MATHEMATICAL_ALPHABETIC_SYMBOLS)
UBLOCKCODE(CHAKMA)
UBLOCKCODE(MEETEI_MAYEK_EXTENSIONS)
UBLOCKCODE(MEROITIC_CURSIVE)
UBLOCKCODE(MEROITIC_HIEROGLYPHS)
UBLOCKCODE(MIAO)
UBLOCKCODE(SHARADA)
UBLOCKCODE(SORA_SOMPENG)
UBLOCKCODE(SUNDANESE_SUPPLEMENT)
UBLOCKCODE(TAKRI)
#if U_ICU_VERSION_MAJOR_NUM >= 54
UBLOCKCODE(BASSA_VAH)
UBLOCKCODE(CAUCASIAN_ALBANIAN)
UBLOCKCODE(COPTIC_EPACT_NUMBERS)
UBLOCKCODE(COMBINING_DIACRITICAL_MARKS_EXTENDED)
UBLOCKCODE(DUPLOYAN)
UBLOCKCODE(ELBASAN)
UBLOCKCODE(GEOMETRIC_SHAPES_EXTENDED)
UBLOCKCODE(GRANTHA)
UBLOCKCODE(KHOJKI)
UBLOCKCODE(KHUDAWADI)
UBLOCKCODE(LATIN_EXTENDED_E)
UBLOCKCODE(LINEAR_A)
UBLOCKCODE(MAHAJANI)
UBLOCKCODE(MANICHAEAN)
UBLOCKCODE(MENDE_KIKAKUI)
UBLOCKCODE(MODI)
UBLOCKCODE(MRO)
UBLOCKCODE(MYANMAR_EXTENDED_B)
UBLOCKCODE(NABATAEAN)
UBLOCKCODE(OLD_NORTH_ARABIAN)
UBLOCKCODE(OLD_PERMIC)
UBLOCKCODE(ORNAMENTAL_DINGBATS)
UBLOCKCODE(PAHAWH_HMONG)
UBLOCKCODE(PALMYRENE)
UBLOCKCODE(PAU_CIN_HAU)
UBLOCKCODE(PSALTER_PAHLAVI)
UBLOCKCODE(SHORTHAND_FORMAT_CONTROLS)
UBLOCKCODE(SIDDHAM)
UBLOCKCODE(SINHALA_ARCHAIC_NUMBERS)
UBLOCKCODE(SUPPLEMENTAL_ARROWS_C)
UBLOCKCODE(TIRHUTA)
UBLOCKCODE(WARANG_CITI)
#endif /* ICU >= 54 */
UBLOCKCODE(COUNT)
UBLOCKCODE(INVALID_CODE)
