/* Generate from http://icu-project.org/apiref/icu4c/uchar_8h.html#a6a2dbc531efce8d77fdb4c314e7fc25e */
UCHARCATEGORY(UNASSIGNED)
UCHARCATEGORY(GENERAL_OTHER_TYPES)
UCHARCATEGORY(UPPERCASE_LETTER)
UCHARCATEGORY(LOWERCASE_LETTER)
UCHARCATEGORY(TITLECASE_LETTER)
UCHARCATEGORY(MODIFIER_LETTER)
UCHARCATEGORY(OTHER_LETTER)
UCHARCATEGORY(NON_SPACING_MARK)
UCHARCATEGORY(ENCLOSING_MARK)
UCHARCATEGORY(COMBINING_SPACING_MARK)
UCHARCATEGORY(DECIMAL_DIGIT_NUMBER)
UCHARCATEGORY(LETTER_NUMBER)
UCHARCATEGORY(OTHER_NUMBER)
UCHARCATEGORY(SPACE_SEPARATOR)
UCHARCATEGORY(LINE_SEPARATOR)
UCHARCATEGORY(PARAGRAPH_SEPARATOR)
UCHARCATEGORY(CONTROL_CHAR)
UCHARCATEGORY(FORMAT_CHAR)
UCHARCATEGORY(PRIVATE_USE_CHAR)
UCHARCATEGORY(SURROGATE)
UCHARCATEGORY(DASH_PUNCTUATION)
UCHARCATEGORY(START_PUNCTUATION)
UCHARCATEGORY(END_PUNCTUATION)
UCHARCATEGORY(CONNECTOR_PUNCTUATION)
UCHARCATEGORY(OTHER_PUNCTUATION)
UCHARCATEGORY(MATH_SYMBOL)
UCHARCATEGORY(CURRENCY_SYMBOL)
UCHARCATEGORY(MODIFIER_SYMBOL)
UCHARCATEGORY(OTHER_SYMBOL)
UCHARCATEGORY(INITIAL_PUNCTUATION)
UCHARCATEGORY(FINAL_PUNCTUATION)
UCHARCATEGORY(CHAR_CATEGORY_COUNT)
