	{ "acdt",  1,  37800, "Australia/Adelaide"            },
	{ "acdt",  1,  37800, "Australia/Broken_Hill"         },
	{ "acdt",  1,  37800, "Australia/Darwin"              },
	{ "acdt",  1,  37800, "Australia/North"               },
	{ "acdt",  1,  37800, "Australia/South"               },
	{ "acdt",  1,  37800, "Australia/Yancowinna"          },
	{ "acst",  0,  34200, "Australia/Adelaide"            },
	{ "acst",  0,  34200, "Australia/Broken_Hill"         },
	{ "acst",  0,  34200, "Australia/Darwin"              },
	{ "acst",  0,  34200, "Australia/North"               },
	{ "acst",  0,  34200, "Australia/South"               },
	{ "acst",  0,  34200, "Australia/Yancowinna"          },
	{ "addt",  1,  -7200, "America/Goose_Bay"             },
	{ "addt",  1,  -7200, "America/Pangnirtung"           },
	{ "adt",   1, -10800, "America/Halifax"               },
	{ "adt",   1, -10800, "America/Barbados"              },
	{ "adt",   1, -10800, "America/Blanc-Sablon"          },
	{ "adt",   1, -10800, "America/Glace_Bay"             },
	{ "adt",   1, -10800, "America/Goose_Bay"             },
	{ "adt",   1, -10800, "America/Martinique"            },
	{ "adt",   1, -10800, "America/Moncton"               },
	{ "adt",   1, -10800, "America/Pangnirtung"           },
	{ "adt",   1, -10800, "America/Thule"                 },
	{ "adt",   1, -10800, "Atlantic/Bermuda"              },
	{ "adt",   1, -10800, "Canada/Atlantic"               },
	{ "aedt",  1,  39600, "Australia/Melbourne"           },
	{ "aedt",  1,  39600, "Antarctica/Macquarie"          },
	{ "aedt",  1,  39600, "Australia/ACT"                 },
	{ "aedt",  1,  39600, "Australia/Brisbane"            },
	{ "aedt",  1,  39600, "Australia/Canberra"            },
	{ "aedt",  1,  39600, "Australia/Currie"              },
	{ "aedt",  1,  39600, "Australia/Hobart"              },
	{ "aedt",  1,  39600, "Australia/Lindeman"            },
	{ "aedt",  1,  39600, "Australia/NSW"                 },
	{ "aedt",  1,  39600, "Australia/Queensland"          },
	{ "aedt",  1,  39600, "Australia/Sydney"              },
	{ "aedt",  1,  39600, "Australia/Tasmania"            },
	{ "aedt",  1,  39600, "Australia/Victoria"            },
	{ "aest",  0,  36000, "Australia/Melbourne"           },
	{ "aest",  0,  36000, "Antarctica/Macquarie"          },
	{ "aest",  0,  36000, "Australia/ACT"                 },
	{ "aest",  0,  36000, "Australia/Brisbane"            },
	{ "aest",  0,  36000, "Australia/Canberra"            },
	{ "aest",  0,  36000, "Australia/Currie"              },
	{ "aest",  0,  36000, "Australia/Hobart"              },
	{ "aest",  0,  36000, "Australia/LHI"                 },
	{ "aest",  0,  36000, "Australia/Lindeman"            },
	{ "aest",  0,  36000, "Australia/Lord_Howe"           },
	{ "aest",  0,  36000, "Australia/NSW"                 },
	{ "aest",  0,  36000, "Australia/Queensland"          },
	{ "aest",  0,  36000, "Australia/Sydney"              },
	{ "aest",  0,  36000, "Australia/Tasmania"            },
	{ "aest",  0,  36000, "Australia/Victoria"            },
	{ "ahdt",  1, -32400, "America/Anchorage"             },
	{ "ahst",  0, -36000, "America/Anchorage"             },
	{ "ahst",  0, -36000, "America/Adak"                  },
	{ "ahst",  0, -36000, "America/Atka"                  },
	{ "akdt",  1, -28800, "America/Anchorage"             },
	{ "akdt",  1, -28800, "America/Juneau"                },
	{ "akdt",  1, -28800, "America/Metlakatla"            },
	{ "akdt",  1, -28800, "America/Nome"                  },
	{ "akdt",  1, -28800, "America/Sitka"                 },
	{ "akdt",  1, -28800, "America/Yakutat"               },
	{ "akst",  0, -32400, "America/Anchorage"             },
	{ "akst",  0, -32400, "America/Juneau"                },
	{ "akst",  0, -32400, "America/Metlakatla"            },
	{ "akst",  0, -32400, "America/Nome"                  },
	{ "akst",  0, -32400, "America/Sitka"                 },
	{ "akst",  0, -32400, "America/Yakutat"               },
	{ "amt",   0, -13840, "America/Asuncion"              },
	{ "amt",   0,   1172, "Europe/Amsterdam"              },
	{ "amt",   0,   5692, "Europe/Athens"                 },
	{ "apt",   1, -10800, "America/Halifax"               },
	{ "apt",   1, -10800, "America/Blanc-Sablon"          },
	{ "apt",   1, -10800, "America/Glace_Bay"             },
	{ "apt",   1, -10800, "America/Moncton"               },
	{ "apt",   1, -10800, "America/Pangnirtung"           },
	{ "apt",   1, -10800, "America/Puerto_Rico"           },
	{ "apt",   1, -10800, "Canada/Atlantic"               },
	{ "apt",   1, -32400, "America/Anchorage"             },
	{ "ast",   0, -14400, "America/Anguilla"              },
	{ "ast",   0, -14400, "America/Antigua"               },
	{ "ast",   0, -14400, "America/Aruba"                 },
	{ "ast",   0, -14400, "America/Barbados"              },
	{ "ast",   0, -14400, "America/Blanc-Sablon"          },
	{ "ast",   0, -14400, "America/Curacao"               },
	{ "ast",   0, -14400, "America/Dominica"              },
	{ "ast",   0, -14400, "America/Glace_Bay"             },
	{ "ast",   0, -14400, "America/Goose_Bay"             },
	{ "ast",   0, -14400, "America/Grand_Turk"            },
	{ "ast",   0, -14400, "America/Grenada"               },
	{ "ast",   0, -14400, "America/Guadeloupe"            },
	{ "ast",   0, -14400, "America/Halifax"               },
	{ "ast",   0, -14400, "America/Kralendijk"            },
	{ "ast",   0, -14400, "America/Lower_Princes"         },
	{ "ast",   0, -14400, "America/Marigot"               },
	{ "ast",   0, -14400, "America/Martinique"            },
	{ "ast",   0, -14400, "America/Miquelon"              },
	{ "ast",   0, -14400, "America/Moncton"               },
	{ "ast",   0, -14400, "America/Montserrat"            },
	{ "ast",   0, -14400, "America/Pangnirtung"           },
	{ "ast",   0, -14400, "America/Port_of_Spain"         },
	{ "ast",   0, -14400, "America/Puerto_Rico"           },
	{ "ast",   0, -14400, "America/Santo_Domingo"         },
	{ "ast",   0, -14400, "America/St_Barthelemy"         },
	{ "ast",   0, -14400, "America/St_Kitts"              },
	{ "ast",   0, -14400, "America/St_Lucia"              },
	{ "ast",   0, -14400, "America/St_Thomas"             },
	{ "ast",   0, -14400, "America/St_Vincent"            },
	{ "ast",   0, -14400, "America/Thule"                 },
	{ "ast",   0, -14400, "America/Tortola"               },
	{ "ast",   0, -14400, "America/Virgin"                },
	{ "ast",   0, -14400, "Atlantic/Bermuda"              },
	{ "ast",   0, -14400, "Canada/Atlantic"               },
	{ "ast",   0, -36000, "America/Anchorage"             },
	{ "awdt",  1,  32400, "Australia/Perth"               },
	{ "awdt",  1,  32400, "Australia/West"                },
	{ "awst",  0,  28800, "Australia/Perth"               },
	{ "awst",  0,  28800, "Australia/West"                },
	{ "awt",   1, -10800, "America/Halifax"               },
	{ "awt",   1, -10800, "America/Blanc-Sablon"          },
	{ "awt",   1, -10800, "America/Glace_Bay"             },
	{ "awt",   1, -10800, "America/Moncton"               },
	{ "awt",   1, -10800, "America/Pangnirtung"           },
	{ "awt",   1, -10800, "America/Puerto_Rico"           },
	{ "awt",   1, -10800, "Canada/Atlantic"               },
	{ "awt",   1, -32400, "America/Anchorage"             },
	{ "bdst",  1,   7200, "Europe/London"                 },
	{ "bdst",  1,   7200, "Europe/Belfast"                },
	{ "bdst",  1,   7200, "Europe/Gibraltar"              },
	{ "bdst",  1,   7200, "Europe/Guernsey"               },
	{ "bdst",  1,   7200, "Europe/Isle_of_Man"            },
	{ "bdst",  1,   7200, "Europe/Jersey"                 },
	{ "bdst",  1,   7200, "GB"                            },
	{ "bdt",   1, -36000, "America/Adak"                  },
	{ "bdt",   1, -36000, "America/Atka"                  },
	{ "bdt",   1, -36000, "America/Nome"                  },
	{ "bmt",   0, -14309, "America/Barbados"              },
	{ "bmt",   0,   6264, "Europe/Tiraspol"               },
	{ "bmt",   0, -17776, "America/Bogota"                },
	{ "bmt",   0,  10656, "Asia/Baghdad"                  },
	{ "bmt",   0,  24124, "Asia/Bangkok"                  },
	{ "bmt",   0,  24124, "Asia/Phnom_Penh"               },
	{ "bmt",   0,  24124, "Asia/Vientiane"                },
	{ "bmt",   0,  25632, "Asia/Jakarta"                  },
	{ "bmt",   0,   6264, "Europe/Bucharest"              },
	{ "bmt",   0,   6264, "Europe/Chisinau"               },
	{ "bst",   1,   3600, "Europe/London"                 },
	{ "bst",   0,   3600, "Europe/London"                 },
	{ "bst",   1, -12756, "America/La_Paz"                },
	{ "bst",   0, -39600, "America/Adak"                  },
	{ "bst",   0, -39600, "America/Atka"                  },
	{ "bst",   0, -39600, "America/Nome"                  },
	{ "bst",   0,   3600, "Europe/Belfast"                },
	{ "bst",   0,   3600, "Europe/Guernsey"               },
	{ "bst",   0,   3600, "Europe/Isle_of_Man"            },
	{ "bst",   0,   3600, "Europe/Jersey"                 },
	{ "bst",   0,   3600, "GB"                            },
	{ "bst",   1,   3600, "Europe/Belfast"                },
	{ "bst",   1,   3600, "Europe/Dublin"                 },
	{ "bst",   1,   3600, "Europe/Gibraltar"              },
	{ "bst",   1,   3600, "Europe/Guernsey"               },
	{ "bst",   1,   3600, "Europe/Isle_of_Man"            },
	{ "bst",   1,   3600, "Europe/Jersey"                 },
	{ "bst",   1,   3600, "GB"                            },
	{ "cast",  0,  34200, "Australia/Adelaide"            },
	{ "cast",  1,  10800, "Africa/Juba"                   },
	{ "cast",  1,  10800, "Africa/Khartoum"               },
	{ "cat",   0,   7200, "Africa/Khartoum"               },
	{ "cat",   0,   7200, "Africa/Blantyre"               },
	{ "cat",   0,   7200, "Africa/Bujumbura"              },
	{ "cat",   0,   7200, "Africa/Gaborone"               },
	{ "cat",   0,   7200, "Africa/Harare"                 },
	{ "cat",   0,   7200, "Africa/Juba"                   },
	{ "cat",   0,   7200, "Africa/Kigali"                 },
	{ "cat",   0,   7200, "Africa/Lubumbashi"             },
	{ "cat",   0,   7200, "Africa/Lusaka"                 },
	{ "cat",   0,   7200, "Africa/Maputo"                 },
	{ "cat",   0,   7200, "Africa/Windhoek"               },
	{ "cddt",  1, -14400, "America/Rankin_Inlet"          },
	{ "cddt",  1, -14400, "America/Resolute"              },
	{ "cdt",   1, -18000, "America/Chicago"               },
	{ "cdt",   1,  32400, "Asia/Shanghai"                 },
	{ "cdt",   1, -14400, "America/Havana"                },
	{ "cdt",   1, -18000, "America/Atikokan"              },
	{ "cdt",   1, -18000, "America/Bahia_Banderas"        },
	{ "cdt",   1, -18000, "America/Belize"                },
	{ "cdt",   1, -18000, "America/Cambridge_Bay"         },
	{ "cdt",   1, -18000, "America/Cancun"                },
	{ "cdt",   1, -18000, "America/Chihuahua"             },
	{ "cdt",   1, -18000, "America/Coral_Harbour"         },
	{ "cdt",   1, -18000, "America/Costa_Rica"            },
	{ "cdt",   1, -18000, "America/El_Salvador"           },
	{ "cdt",   1, -18000, "America/Fort_Wayne"            },
	{ "cdt",   1, -18000, "America/Guatemala"             },
	{ "cdt",   1, -18000, "America/Indiana/Indianapolis"  },
	{ "cdt",   1, -18000, "America/Indiana/Knox"          },
	{ "cdt",   1, -18000, "America/Indiana/Marengo"       },
	{ "cdt",   1, -18000, "America/Indiana/Petersburg"    },
	{ "cdt",   1, -18000, "America/Indiana/Tell_City"     },
	{ "cdt",   1, -18000, "America/Indiana/Vevay"         },
	{ "cdt",   1, -18000, "America/Indiana/Vincennes"     },
	{ "cdt",   1, -18000, "America/Indiana/Winamac"       },
	{ "cdt",   1, -18000, "America/Indianapolis"          },
	{ "cdt",   1, -18000, "America/Iqaluit"               },
	{ "cdt",   1, -18000, "America/Kentucky/Louisville"   },
	{ "cdt",   1, -18000, "America/Kentucky/Monticello"   },
	{ "cdt",   1, -18000, "America/Knox_IN"               },
	{ "cdt",   1, -18000, "America/Louisville"            },
	{ "cdt",   1, -18000, "America/Managua"               },
	{ "cdt",   1, -18000, "America/Matamoros"             },
	{ "cdt",   1, -18000, "America/Menominee"             },
	{ "cdt",   1, -18000, "America/Merida"                },
	{ "cdt",   1, -18000, "America/Mexico_City"           },
	{ "cdt",   1, -18000, "America/Monterrey"             },
	{ "cdt",   1, -18000, "America/North_Dakota/Beulah"   },
	{ "cdt",   1, -18000, "America/North_Dakota/Center"   },
	{ "cdt",   1, -18000, "America/North_Dakota/New_Salem" },
	{ "cdt",   1, -18000, "America/Ojinaga"               },
	{ "cdt",   1, -18000, "America/Pangnirtung"           },
	{ "cdt",   1, -18000, "America/Rainy_River"           },
	{ "cdt",   1, -18000, "America/Rankin_Inlet"          },
	{ "cdt",   1, -18000, "America/Resolute"              },
	{ "cdt",   1, -18000, "America/Tegucigalpa"           },
	{ "cdt",   1, -18000, "America/Winnipeg"              },
	{ "cdt",   1, -18000, "Canada/Central"                },
	{ "cdt",   1, -18000, "Mexico/General"                },
	{ "cdt",   1,  32400, "Asia/Chongqing"                },
	{ "cdt",   1,  32400, "Asia/Chungking"                },
	{ "cdt",   1,  32400, "Asia/Harbin"                   },
	{ "cdt",   1,  32400, "Asia/Macao"                    },
	{ "cdt",   1,  32400, "Asia/Macau"                    },
	{ "cdt",   1,  32400, "Asia/Taipei"                   },
	{ "cdt",   1,  32400, "PRC"                           },
	{ "cdt",   1,  32400, "ROC"                           },
	{ "cemt",  1,  10800, "Europe/Berlin"                 },
	{ "cest",  1,   7200, "Europe/Berlin"                 },
	{ "cest",  1,  10800, "Europe/Kaliningrad"            },
	{ "cest",  1,   7200, "Africa/Algiers"                },
	{ "cest",  1,   7200, "Africa/Ceuta"                  },
	{ "cest",  1,   7200, "Africa/Tripoli"                },
	{ "cest",  1,   7200, "Africa/Tunis"                  },
	{ "cest",  1,   7200, "Arctic/Longyearbyen"           },
	{ "cest",  1,   7200, "Atlantic/Jan_Mayen"            },
	{ "cest",  1,   7200, "Europe/Amsterdam"              },
	{ "cest",  1,   7200, "Europe/Andorra"                },
	{ "cest",  1,   7200, "Europe/Athens"                 },
	{ "cest",  1,   7200, "Europe/Belgrade"               },
	{ "cest",  1,   7200, "Europe/Bratislava"             },
	{ "cest",  1,   7200, "Europe/Brussels"               },
	{ "cest",  1,   7200, "Europe/Budapest"               },
	{ "cest",  1,   7200, "Europe/Busingen"               },
	{ "cest",  1,   7200, "Europe/Chisinau"               },
	{ "cest",  1,   7200, "Europe/Copenhagen"             },
	{ "cest",  1,   7200, "Europe/Gibraltar"              },
	{ "cest",  1,   7200, "Europe/Kaliningrad"            },
	{ "cest",  1,   7200, "Europe/Kiev"                   },
	{ "cest",  1,   7200, "Europe/Lisbon"                 },
	{ "cest",  1,   7200, "Europe/Ljubljana"              },
	{ "cest",  1,   7200, "Europe/Luxembourg"             },
	{ "cest",  1,   7200, "Europe/Madrid"                 },
	{ "cest",  1,   7200, "Europe/Malta"                  },
	{ "cest",  1,   7200, "Europe/Minsk"                  },
	{ "cest",  1,   7200, "Europe/Monaco"                 },
	{ "cest",  1,   7200, "Europe/Oslo"                   },
	{ "cest",  1,   7200, "Europe/Paris"                  },
	{ "cest",  1,   7200, "Europe/Podgorica"              },
	{ "cest",  1,   7200, "Europe/Prague"                 },
	{ "cest",  1,   7200, "Europe/Riga"                   },
	{ "cest",  1,   7200, "Europe/Rome"                   },
	{ "cest",  1,   7200, "Europe/San_Marino"             },
	{ "cest",  1,   7200, "Europe/Sarajevo"               },
	{ "cest",  1,   7200, "Europe/Simferopol"             },
	{ "cest",  1,   7200, "Europe/Skopje"                 },
	{ "cest",  1,   7200, "Europe/Sofia"                  },
	{ "cest",  1,   7200, "Europe/Stockholm"              },
	{ "cest",  1,   7200, "Europe/Tallinn"                },
	{ "cest",  1,   7200, "Europe/Tirane"                 },
	{ "cest",  1,   7200, "Europe/Tiraspol"               },
	{ "cest",  1,   7200, "Europe/Uzhgorod"               },
	{ "cest",  1,   7200, "Europe/Vaduz"                  },
	{ "cest",  1,   7200, "Europe/Vatican"                },
	{ "cest",  1,   7200, "Europe/Vienna"                 },
	{ "cest",  1,   7200, "Europe/Vilnius"                },
	{ "cest",  1,   7200, "Europe/Warsaw"                 },
	{ "cest",  1,   7200, "Europe/Zagreb"                 },
	{ "cest",  1,   7200, "Europe/Zaporozhye"             },
	{ "cest",  1,   7200, "Europe/Zurich"                 },
	{ "cet",   0,   3600, "Europe/Berlin"                 },
	{ "cet",   0,   7200, "Europe/Kaliningrad"            },
	{ "cet",   0,   3600, "Africa/Algiers"                },
	{ "cet",   0,   3600, "Africa/Ceuta"                  },
	{ "cet",   0,   3600, "Africa/Tripoli"                },
	{ "cet",   0,   3600, "Africa/Tunis"                  },
	{ "cet",   0,   3600, "Arctic/Longyearbyen"           },
	{ "cet",   0,   3600, "Atlantic/Jan_Mayen"            },
	{ "cet",   0,   3600, "Europe/Amsterdam"              },
	{ "cet",   0,   3600, "Europe/Andorra"                },
	{ "cet",   0,   3600, "Europe/Athens"                 },
	{ "cet",   0,   3600, "Europe/Belgrade"               },
	{ "cet",   0,   3600, "Europe/Bratislava"             },
	{ "cet",   0,   3600, "Europe/Brussels"               },
	{ "cet",   0,   3600, "Europe/Budapest"               },
	{ "cet",   0,   3600, "Europe/Busingen"               },
	{ "cet",   0,   3600, "Europe/Chisinau"               },
	{ "cet",   0,   3600, "Europe/Copenhagen"             },
	{ "cet",   0,   3600, "Europe/Gibraltar"              },
	{ "cet",   0,   3600, "Europe/Kaliningrad"            },
	{ "cet",   0,   3600, "Europe/Kiev"                   },
	{ "cet",   0,   3600, "Europe/Lisbon"                 },
	{ "cet",   0,   3600, "Europe/Ljubljana"              },
	{ "cet",   0,   3600, "Europe/Luxembourg"             },
	{ "cet",   0,   3600, "Europe/Madrid"                 },
	{ "cet",   0,   3600, "Europe/Malta"                  },
	{ "cet",   0,   3600, "Europe/Minsk"                  },
	{ "cet",   0,   3600, "Europe/Monaco"                 },
	{ "cet",   0,   3600, "Europe/Oslo"                   },
	{ "cet",   0,   3600, "Europe/Paris"                  },
	{ "cet",   0,   3600, "Europe/Podgorica"              },
	{ "cet",   0,   3600, "Europe/Prague"                 },
	{ "cet",   0,   3600, "Europe/Riga"                   },
	{ "cet",   0,   3600, "Europe/Rome"                   },
	{ "cet",   0,   3600, "Europe/San_Marino"             },
	{ "cet",   0,   3600, "Europe/Sarajevo"               },
	{ "cet",   0,   3600, "Europe/Simferopol"             },
	{ "cet",   0,   3600, "Europe/Skopje"                 },
	{ "cet",   0,   3600, "Europe/Sofia"                  },
	{ "cet",   0,   3600, "Europe/Stockholm"              },
	{ "cet",   0,   3600, "Europe/Tallinn"                },
	{ "cet",   0,   3600, "Europe/Tirane"                 },
	{ "cet",   0,   3600, "Europe/Tiraspol"               },
	{ "cet",   0,   3600, "Europe/Uzhgorod"               },
	{ "cet",   0,   3600, "Europe/Vaduz"                  },
	{ "cet",   0,   3600, "Europe/Vatican"                },
	{ "cet",   0,   3600, "Europe/Vienna"                 },
	{ "cet",   0,   3600, "Europe/Vilnius"                },
	{ "cet",   0,   3600, "Europe/Warsaw"                 },
	{ "cet",   0,   3600, "Europe/Zagreb"                 },
	{ "cet",   0,   3600, "Europe/Zaporozhye"             },
	{ "cet",   0,   3600, "Europe/Zurich"                 },
	{ "cmt",   0, -15408, "America/Argentina/Buenos_Aires" },
	{ "cmt",   0, -15408, "America/Argentina/Catamarca"   },
	{ "cmt",   0, -15408, "America/Argentina/ComodRivadavia" },
	{ "cmt",   0, -15408, "America/Argentina/Cordoba"     },
	{ "cmt",   0, -15408, "America/Argentina/Jujuy"       },
	{ "cmt",   0, -15408, "America/Argentina/La_Rioja"    },
	{ "cmt",   0, -15408, "America/Argentina/Mendoza"     },
	{ "cmt",   0, -15408, "America/Argentina/Rio_Gallegos" },
	{ "cmt",   0, -15408, "America/Argentina/Salta"       },
	{ "cmt",   0, -15408, "America/Argentina/San_Juan"    },
	{ "cmt",   0, -15408, "America/Argentina/San_Luis"    },
	{ "cmt",   0, -15408, "America/Argentina/Tucuman"     },
	{ "cmt",   0, -15408, "America/Argentina/Ushuaia"     },
	{ "cmt",   0, -15408, "America/Buenos_Aires"          },
	{ "cmt",   0, -15408, "America/Catamarca"             },
	{ "cmt",   0, -15408, "America/Cordoba"               },
	{ "cmt",   0, -15408, "America/Jujuy"                 },
	{ "cmt",   0, -15408, "America/Mendoza"               },
	{ "cmt",   0, -15408, "America/Rosario"               },
	{ "cmt",   0, -16060, "America/Caracas"               },
	{ "cmt",   0, -16356, "America/La_Paz"                },
	{ "cmt",   0, -19176, "America/Cayman"                },
	{ "cmt",   0, -19176, "America/Panama"                },
	{ "cmt",   0,   6900, "Europe/Chisinau"               },
	{ "cmt",   0,   6900, "Europe/Tiraspol"               },
	{ "cpt",   1, -18000, "America/Chicago"               },
	{ "cpt",   1, -18000, "America/Atikokan"              },
	{ "cpt",   1, -18000, "America/Coral_Harbour"         },
	{ "cpt",   1, -18000, "America/Fort_Wayne"            },
	{ "cpt",   1, -18000, "America/Indiana/Indianapolis"  },
	{ "cpt",   1, -18000, "America/Indiana/Knox"          },
	{ "cpt",   1, -18000, "America/Indiana/Marengo"       },
	{ "cpt",   1, -18000, "America/Indiana/Petersburg"    },
	{ "cpt",   1, -18000, "America/Indiana/Tell_City"     },
	{ "cpt",   1, -18000, "America/Indiana/Vevay"         },
	{ "cpt",   1, -18000, "America/Indiana/Vincennes"     },
	{ "cpt",   1, -18000, "America/Indiana/Winamac"       },
	{ "cpt",   1, -18000, "America/Indianapolis"          },
	{ "cpt",   1, -18000, "America/Kentucky/Louisville"   },
	{ "cpt",   1, -18000, "America/Kentucky/Monticello"   },
	{ "cpt",   1, -18000, "America/Knox_IN"               },
	{ "cpt",   1, -18000, "America/Louisville"            },
	{ "cpt",   1, -18000, "America/Menominee"             },
	{ "cpt",   1, -18000, "America/Rainy_River"           },
	{ "cpt",   1, -18000, "America/Winnipeg"              },
	{ "cpt",   1, -18000, "Canada/Central"                },
	{ "cst",   0, -21600, "America/Chicago"               },
	{ "cst",   0, -18000, "America/Havana"                },
	{ "cst",   0, -21600, "America/Atikokan"              },
	{ "cst",   0, -21600, "America/Bahia_Banderas"        },
	{ "cst",   0, -21600, "America/Belize"                },
	{ "cst",   0, -21600, "America/Cambridge_Bay"         },
	{ "cst",   0, -21600, "America/Cancun"                },
	{ "cst",   0, -21600, "America/Chihuahua"             },
	{ "cst",   0, -21600, "America/Coral_Harbour"         },
	{ "cst",   0, -21600, "America/Costa_Rica"            },
	{ "cst",   0, -21600, "America/Detroit"               },
	{ "cst",   0, -21600, "America/El_Salvador"           },
	{ "cst",   0, -21600, "America/Fort_Wayne"            },
	{ "cst",   0, -21600, "America/Guatemala"             },
	{ "cst",   0, -21600, "America/Hermosillo"            },
	{ "cst",   0, -21600, "America/Indiana/Indianapolis"  },
	{ "cst",   0, -21600, "America/Indiana/Knox"          },
	{ "cst",   0, -21600, "America/Indiana/Marengo"       },
	{ "cst",   0, -21600, "America/Indiana/Petersburg"    },
	{ "cst",   0, -21600, "America/Indiana/Tell_City"     },
	{ "cst",   0, -21600, "America/Indiana/Vevay"         },
	{ "cst",   0, -21600, "America/Indiana/Vincennes"     },
	{ "cst",   0, -21600, "America/Indiana/Winamac"       },
	{ "cst",   0, -21600, "America/Indianapolis"          },
	{ "cst",   0, -21600, "America/Iqaluit"               },
	{ "cst",   0, -21600, "America/Kentucky/Louisville"   },
	{ "cst",   0, -21600, "America/Kentucky/Monticello"   },
	{ "cst",   0, -21600, "America/Knox_IN"               },
	{ "cst",   0, -21600, "America/Louisville"            },
	{ "cst",   0, -21600, "America/Managua"               },
	{ "cst",   0, -21600, "America/Matamoros"             },
	{ "cst",   0, -21600, "America/Mazatlan"              },
	{ "cst",   0, -21600, "America/Menominee"             },
	{ "cst",   0, -21600, "America/Merida"                },
	{ "cst",   0, -21600, "America/Mexico_City"           },
	{ "cst",   0, -21600, "America/Monterrey"             },
	{ "cst",   0, -21600, "America/North_Dakota/Beulah"   },
	{ "cst",   0, -21600, "America/North_Dakota/Center"   },
	{ "cst",   0, -21600, "America/North_Dakota/New_Salem" },
	{ "cst",   0, -21600, "America/Ojinaga"               },
	{ "cst",   0, -21600, "America/Pangnirtung"           },
	{ "cst",   0, -21600, "America/Rainy_River"           },
	{ "cst",   0, -21600, "America/Rankin_Inlet"          },
	{ "cst",   0, -21600, "America/Regina"                },
	{ "cst",   0, -21600, "America/Resolute"              },
	{ "cst",   0, -21600, "America/Swift_Current"         },
	{ "cst",   0, -21600, "America/Tegucigalpa"           },
	{ "cst",   0, -21600, "America/Thunder_Bay"           },
	{ "cst",   0, -21600, "America/Winnipeg"              },
	{ "cst",   0, -21600, "Canada/Central"                },
	{ "cst",   0, -21600, "Canada/Saskatchewan"           },
	{ "cst",   0, -21600, "Mexico/BajaSur"                },
	{ "cst",   0, -21600, "Mexico/General"                },
	{ "cst",   0,  28800, "Asia/Chongqing"                },
	{ "cst",   0,  28800, "Asia/Chungking"                },
	{ "cst",   0,  28800, "Asia/Harbin"                   },
	{ "cst",   0,  28800, "Asia/Macao"                    },
	{ "cst",   0,  28800, "Asia/Macau"                    },
	{ "cst",   0,  28800, "Asia/Shanghai"                 },
	{ "cst",   0,  28800, "Asia/Taipei"                   },
	{ "cst",   0,  28800, "PRC"                           },
	{ "cst",   0,  28800, "ROC"                           },
	{ "cwt",   1, -18000, "America/Chicago"               },
	{ "cwt",   1, -18000, "America/Atikokan"              },
	{ "cwt",   1, -18000, "America/Coral_Harbour"         },
	{ "cwt",   1, -18000, "America/Fort_Wayne"            },
	{ "cwt",   1, -18000, "America/Indiana/Indianapolis"  },
	{ "cwt",   1, -18000, "America/Indiana/Knox"          },
	{ "cwt",   1, -18000, "America/Indiana/Marengo"       },
	{ "cwt",   1, -18000, "America/Indiana/Petersburg"    },
	{ "cwt",   1, -18000, "America/Indiana/Tell_City"     },
	{ "cwt",   1, -18000, "America/Indiana/Vevay"         },
	{ "cwt",   1, -18000, "America/Indiana/Vincennes"     },
	{ "cwt",   1, -18000, "America/Indiana/Winamac"       },
	{ "cwt",   1, -18000, "America/Indianapolis"          },
	{ "cwt",   1, -18000, "America/Kentucky/Louisville"   },
	{ "cwt",   1, -18000, "America/Kentucky/Monticello"   },
	{ "cwt",   1, -18000, "America/Knox_IN"               },
	{ "cwt",   1, -18000, "America/Louisville"            },
	{ "cwt",   1, -18000, "America/Menominee"             },
	{ "cwt",   1, -18000, "America/Mexico_City"           },
	{ "cwt",   1, -18000, "America/Rainy_River"           },
	{ "cwt",   1, -18000, "America/Winnipeg"              },
	{ "cwt",   1, -18000, "Canada/Central"                },
	{ "cwt",   1, -18000, "Mexico/General"                },
	{ "chst",  0,  36000, "Pacific/Guam"                  },
	{ "chst",  0,  36000, "Pacific/Saipan"                },
	{ "dmt",   0,  -1521, "Europe/Dublin"                 },
	{ "eat",   0,  10800, "Africa/Khartoum"               },
	{ "eat",   0,  10800, "Africa/Addis_Ababa"            },
	{ "eat",   0,  10800, "Africa/Asmara"                 },
	{ "eat",   0,  10800, "Africa/Asmera"                 },
	{ "eat",   0,  10800, "Africa/Dar_es_Salaam"          },
	{ "eat",   0,  10800, "Africa/Djibouti"               },
	{ "eat",   0,  10800, "Africa/Juba"                   },
	{ "eat",   0,  10800, "Africa/Kampala"                },
	{ "eat",   0,  10800, "Africa/Mogadishu"              },
	{ "eat",   0,  10800, "Africa/Nairobi"                },
	{ "eat",   0,  10800, "Indian/Antananarivo"           },
	{ "eat",   0,  10800, "Indian/Comoro"                 },
	{ "eat",   0,  10800, "Indian/Mayotte"                },
	{ "eddt",  1, -10800, "America/Iqaluit"               },
	{ "edt",   1, -14400, "America/New_York"              },
	{ "edt",   1, -14400, "America/Cancun"                },
	{ "edt",   1, -14400, "America/Detroit"               },
	{ "edt",   1, -14400, "America/Fort_Wayne"            },
	{ "edt",   1, -14400, "America/Grand_Turk"            },
	{ "edt",   1, -14400, "America/Indiana/Indianapolis"  },
	{ "edt",   1, -14400, "America/Indiana/Marengo"       },
	{ "edt",   1, -14400, "America/Indiana/Petersburg"    },
	{ "edt",   1, -14400, "America/Indiana/Tell_City"     },
	{ "edt",   1, -14400, "America/Indiana/Vevay"         },
	{ "edt",   1, -14400, "America/Indiana/Vincennes"     },
	{ "edt",   1, -14400, "America/Indiana/Winamac"       },
	{ "edt",   1, -14400, "America/Indianapolis"          },
	{ "edt",   1, -14400, "America/Iqaluit"               },
	{ "edt",   1, -14400, "America/Jamaica"               },
	{ "edt",   1, -14400, "America/Kentucky/Louisville"   },
	{ "edt",   1, -14400, "America/Kentucky/Monticello"   },
	{ "edt",   1, -14400, "America/Louisville"            },
	{ "edt",   1, -14400, "America/Montreal"              },
	{ "edt",   1, -14400, "America/Nassau"                },
	{ "edt",   1, -14400, "America/Nipigon"               },
	{ "edt",   1, -14400, "America/Pangnirtung"           },
	{ "edt",   1, -14400, "America/Port-au-Prince"        },
	{ "edt",   1, -14400, "America/Santo_Domingo"         },
	{ "edt",   1, -14400, "America/Thunder_Bay"           },
	{ "edt",   1, -14400, "America/Toronto"               },
	{ "edt",   1, -14400, "Canada/Eastern"                },
	{ "eest",  1,  10800, "Europe/Helsinki"               },
	{ "eest",  1,  10800, "Africa/Cairo"                  },
	{ "eest",  1,  10800, "Asia/Amman"                    },
	{ "eest",  1,  10800, "Asia/Beirut"                   },
	{ "eest",  1,  10800, "Asia/Damascus"                 },
	{ "eest",  1,  10800, "Asia/Famagusta"                },
	{ "eest",  1,  10800, "Asia/Gaza"                     },
	{ "eest",  1,  10800, "Asia/Hebron"                   },
	{ "eest",  1,  10800, "Asia/Istanbul"                 },
	{ "eest",  1,  10800, "Asia/Nicosia"                  },
	{ "eest",  1,  10800, "Europe/Athens"                 },
	{ "eest",  1,  10800, "Europe/Bucharest"              },
	{ "eest",  1,  10800, "Europe/Chisinau"               },
	{ "eest",  1,  10800, "Europe/Istanbul"               },
	{ "eest",  1,  10800, "Europe/Kaliningrad"            },
	{ "eest",  1,  10800, "Europe/Kiev"                   },
	{ "eest",  1,  10800, "Europe/Mariehamn"              },
	{ "eest",  1,  10800, "Europe/Minsk"                  },
	{ "eest",  1,  10800, "Europe/Moscow"                 },
	{ "eest",  1,  10800, "Europe/Nicosia"                },
	{ "eest",  1,  10800, "Europe/Riga"                   },
	{ "eest",  1,  10800, "Europe/Simferopol"             },
	{ "eest",  1,  10800, "Europe/Sofia"                  },
	{ "eest",  1,  10800, "Europe/Tallinn"                },
	{ "eest",  1,  10800, "Europe/Tiraspol"               },
	{ "eest",  1,  10800, "Europe/Uzhgorod"               },
	{ "eest",  1,  10800, "Europe/Vilnius"                },
	{ "eest",  1,  10800, "Europe/Warsaw"                 },
	{ "eest",  1,  10800, "Europe/Zaporozhye"             },
	{ "eet",   0,   7200, "Europe/Helsinki"               },
	{ "eet",   0,   7200, "Africa/Cairo"                  },
	{ "eet",   0,   7200, "Africa/Tripoli"                },
	{ "eet",   0,   7200, "Asia/Amman"                    },
	{ "eet",   0,   7200, "Asia/Beirut"                   },
	{ "eet",   0,   7200, "Asia/Damascus"                 },
	{ "eet",   0,   7200, "Asia/Famagusta"                },
	{ "eet",   0,   7200, "Asia/Gaza"                     },
	{ "eet",   0,   7200, "Asia/Hebron"                   },
	{ "eet",   0,   7200, "Asia/Istanbul"                 },
	{ "eet",   0,   7200, "Asia/Nicosia"                  },
	{ "eet",   0,   7200, "Europe/Athens"                 },
	{ "eet",   0,   7200, "Europe/Bucharest"              },
	{ "eet",   0,   7200, "Europe/Chisinau"               },
	{ "eet",   0,   7200, "Europe/Istanbul"               },
	{ "eet",   0,   7200, "Europe/Kaliningrad"            },
	{ "eet",   0,   7200, "Europe/Kiev"                   },
	{ "eet",   0,   7200, "Europe/Mariehamn"              },
	{ "eet",   0,   7200, "Europe/Minsk"                  },
	{ "eet",   0,   7200, "Europe/Moscow"                 },
	{ "eet",   0,   7200, "Europe/Nicosia"                },
	{ "eet",   0,   7200, "Europe/Riga"                   },
	{ "eet",   0,   7200, "Europe/Simferopol"             },
	{ "eet",   0,   7200, "Europe/Sofia"                  },
	{ "eet",   0,   7200, "Europe/Tallinn"                },
	{ "eet",   0,   7200, "Europe/Tiraspol"               },
	{ "eet",   0,   7200, "Europe/Uzhgorod"               },
	{ "eet",   0,   7200, "Europe/Vilnius"                },
	{ "eet",   0,   7200, "Europe/Warsaw"                 },
	{ "eet",   0,   7200, "Europe/Zaporozhye"             },
	{ "emt",   0, -26248, "Chile/EasterIsland"            },
	{ "emt",   0, -26248, "Pacific/Easter"                },
	{ "ept",   1, -14400, "America/New_York"              },
	{ "ept",   1, -14400, "America/Detroit"               },
	{ "ept",   1, -14400, "America/Iqaluit"               },
	{ "ept",   1, -14400, "America/Montreal"              },
	{ "ept",   1, -14400, "America/Nipigon"               },
	{ "ept",   1, -14400, "America/Thunder_Bay"           },
	{ "ept",   1, -14400, "America/Toronto"               },
	{ "ept",   1, -14400, "Canada/Eastern"                },
	{ "est",   0, -18000, "America/New_York"              },
	{ "est",   0, -18000, "America/Atikokan"              },
	{ "est",   0, -18000, "America/Cambridge_Bay"         },
	{ "est",   0, -18000, "America/Cancun"                },
	{ "est",   0, -18000, "America/Cayman"                },
	{ "est",   0, -18000, "America/Chicago"               },
	{ "est",   0, -18000, "America/Coral_Harbour"         },
	{ "est",   0, -18000, "America/Detroit"               },
	{ "est",   0, -18000, "America/Fort_Wayne"            },
	{ "est",   0, -18000, "America/Grand_Turk"            },
	{ "est",   0, -18000, "America/Indiana/Indianapolis"  },
	{ "est",   0, -18000, "America/Indiana/Knox"          },
	{ "est",   0, -18000, "America/Indiana/Marengo"       },
	{ "est",   0, -18000, "America/Indiana/Petersburg"    },
	{ "est",   0, -18000, "America/Indiana/Tell_City"     },
	{ "est",   0, -18000, "America/Indiana/Vevay"         },
	{ "est",   0, -18000, "America/Indiana/Vincennes"     },
	{ "est",   0, -18000, "America/Indiana/Winamac"       },
	{ "est",   0, -18000, "America/Indianapolis"          },
	{ "est",   0, -18000, "America/Iqaluit"               },
	{ "est",   0, -18000, "America/Jamaica"               },
	{ "est",   0, -18000, "America/Kentucky/Louisville"   },
	{ "est",   0, -18000, "America/Kentucky/Monticello"   },
	{ "est",   0, -18000, "America/Knox_IN"               },
	{ "est",   0, -18000, "America/Louisville"            },
	{ "est",   0, -18000, "America/Managua"               },
	{ "est",   0, -18000, "America/Menominee"             },
	{ "est",   0, -18000, "America/Merida"                },
	{ "est",   0, -18000, "America/Moncton"               },
	{ "est",   0, -18000, "America/Montreal"              },
	{ "est",   0, -18000, "America/Nassau"                },
	{ "est",   0, -18000, "America/Nipigon"               },
	{ "est",   0, -18000, "America/Panama"                },
	{ "est",   0, -18000, "America/Pangnirtung"           },
	{ "est",   0, -18000, "America/Port-au-Prince"        },
	{ "est",   0, -18000, "America/Rankin_Inlet"          },
	{ "est",   0, -18000, "America/Resolute"              },
	{ "est",   0, -18000, "America/Santo_Domingo"         },
	{ "est",   0, -18000, "America/Thunder_Bay"           },
	{ "est",   0, -18000, "America/Toronto"               },
	{ "est",   0, -18000, "Canada/Eastern"                },
	{ "ewt",   1, -14400, "America/New_York"              },
	{ "ewt",   1, -14400, "America/Detroit"               },
	{ "ewt",   1, -14400, "America/Iqaluit"               },
	{ "ewt",   1, -14400, "America/Montreal"              },
	{ "ewt",   1, -14400, "America/Nipigon"               },
	{ "ewt",   1, -14400, "America/Thunder_Bay"           },
	{ "ewt",   1, -14400, "America/Toronto"               },
	{ "ewt",   1, -14400, "Canada/Eastern"                },
	{ "ffmt",  0, -14660, "America/Martinique"            },
	{ "fmt",   0,  -4056, "Atlantic/Madeira"              },
	{ "gdt",   1,  39600, "Pacific/Guam"                  },
	{ "gdt",   1,  39600, "Pacific/Saipan"                },
	{ "gmt",   0,      0, "Africa/Abidjan"                },
	{ "gmt",   0,      0, "Africa/Accra"                  },
	{ "gmt",   0,      0, "Africa/Bamako"                 },
	{ "gmt",   0,      0, "Africa/Banjul"                 },
	{ "gmt",   0,      0, "Africa/Bissau"                 },
	{ "gmt",   0,      0, "Africa/Conakry"                },
	{ "gmt",   0,      0, "Africa/Dakar"                  },
	{ "gmt",   0,      0, "Africa/Freetown"               },
	{ "gmt",   0,      0, "Africa/Lome"                   },
	{ "gmt",   0,      0, "Africa/Monrovia"               },
	{ "gmt",   0,      0, "Africa/Nouakchott"             },
	{ "gmt",   0,      0, "Africa/Ouagadougou"            },
	{ "gmt",   0,      0, "Africa/Sao_Tome"               },
	{ "gmt",   0,      0, "Africa/Timbuktu"               },
	{ "gmt",   0,      0, "America/Danmarkshavn"          },
	{ "gmt",   0,      0, "Atlantic/Reykjavik"            },
	{ "gmt",   0,      0, "Atlantic/St_Helena"            },
	{ "gmt",   0,      0, "Etc/GMT"                       },
	{ "gmt",   0,      0, "Etc/Greenwich"                 },
	{ "gmt",   0,      0, "Europe/Belfast"                },
	{ "gmt",   0,      0, "Europe/Dublin"                 },
	{ "gmt",   0,      0, "Europe/Gibraltar"              },
	{ "gmt",   0,      0, "Europe/Guernsey"               },
	{ "gmt",   0,      0, "Europe/Isle_of_Man"            },
	{ "gmt",   0,      0, "Europe/Jersey"                 },
	{ "gmt",   0,      0, "Europe/London"                 },
	{ "gmt",   0,      0, "GB"                            },
	{ "gmt",   1,      0, "Europe/Bratislava"             },
	{ "gmt",   1,      0, "Europe/Dublin"                 },
	{ "gmt",   1,      0, "Europe/Prague"                 },
	{ "gst",   0,  36000, "Pacific/Guam"                  },
	{ "gst",   0,  36000, "Pacific/Saipan"                },
	{ "hdt",   1, -34200, "Pacific/Honolulu"              },
	{ "hdt",   1, -32400, "America/Adak"                  },
	{ "hdt",   1, -32400, "America/Atka"                  },
	{ "hdt",   1, -34200, "Pacific/Johnston"              },
	{ "hkst",  1,  32400, "Asia/Hong_Kong"                },
	{ "hkt",   0,  28800, "Asia/Hong_Kong"                },
	{ "hkt",   0,  30600, "Asia/Hong_Kong"                },
	{ "hmt",   0, -19776, "America/Havana"                },
	{ "hmt",   0,  -6872, "Atlantic/Azores"               },
	{ "hmt",   0,  21200, "Asia/Dacca"                    },
	{ "hmt",   0,  21200, "Asia/Dhaka"                    },
	{ "hmt",   0,   5989, "Europe/Helsinki"               },
	{ "hmt",   0,   5989, "Europe/Mariehamn"              },
	{ "hpt",   1, -34200, "Pacific/Honolulu"              },
	{ "hpt",   1, -34200, "Pacific/Johnston"              },
	{ "hst",   0, -36000, "Pacific/Honolulu"              },
	{ "hst",   0, -37800, "Pacific/Honolulu"              },
	{ "hst",   0, -36000, "America/Adak"                  },
	{ "hst",   0, -36000, "America/Atka"                  },
	{ "hst",   0, -36000, "Pacific/Johnston"              },
	{ "hst",   0, -37800, "Pacific/Johnston"              },
	{ "hwt",   1, -34200, "Pacific/Honolulu"              },
	{ "hwt",   1, -34200, "Pacific/Johnston"              },
	{ "iddt",  1,  14400, "Asia/Jerusalem"                },
	{ "iddt",  1,  14400, "Asia/Tel_Aviv"                 },
	{ "idt",   1,  10800, "Asia/Jerusalem"                },
	{ "idt",   1,  10800, "Asia/Gaza"                     },
	{ "idt",   1,  10800, "Asia/Hebron"                   },
	{ "idt",   1,  10800, "Asia/Tel_Aviv"                 },
	{ "imt",   0,  25025, "Asia/Irkutsk"                  },
	{ "imt",   0,   7016, "Asia/Istanbul"                 },
	{ "imt",   0,   7016, "Europe/Istanbul"               },
	{ "ist",   0,   7200, "Asia/Jerusalem"                },
	{ "ist",   0,  19800, "Asia/Calcutta"                 },
	{ "ist",   0,  19800, "Asia/Kolkata"                  },
	{ "ist",   1,   2079, "Europe/Dublin"                 },
	{ "ist",   0,   3600, "Europe/Dublin"                 },
	{ "ist",   1,   3600, "Europe/Dublin"                 },
	{ "ist",   0,   7200, "Asia/Gaza"                     },
	{ "ist",   0,   7200, "Asia/Hebron"                   },
	{ "ist",   0,   7200, "Asia/Tel_Aviv"                 },
	{ "jdt",   1,  36000, "Asia/Tokyo"                    },
	{ "jmt",   0,   8440, "Asia/Jerusalem"                },
	{ "jmt",   0,   8440, "Asia/Tel_Aviv"                 },
	{ "jst",   0,  32400, "Asia/Tokyo"                    },
	{ "jst",   0,  32400, "Asia/Hong_Kong"                },
	{ "jst",   0,  32400, "Asia/Manila"                   },
	{ "jst",   0,  32400, "Asia/Pyongyang"                },
	{ "jst",   0,  32400, "Asia/Seoul"                    },
	{ "jst",   0,  32400, "Asia/Taipei"                   },
	{ "jst",   0,  32400, "ROC"                           },
	{ "jst",   0,  32400, "ROK"                           },
	{ "kdt",   1,  36000, "Asia/Seoul"                    },
	{ "kdt",   1,  34200, "Asia/Seoul"                    },
	{ "kdt",   1,  34200, "ROK"                           },
	{ "kdt",   1,  36000, "ROK"                           },
	{ "kmt",   0,   5736, "Europe/Vilnius"                },
	{ "kmt",   0, -18430, "America/Grand_Turk"            },
	{ "kmt",   0, -18430, "America/Jamaica"               },
	{ "kmt",   0,   7324, "Europe/Kiev"                   },
	{ "kst",   0,  30600, "Asia/Seoul"                    },
	{ "kst",   0,  32400, "Asia/Pyongyang"                },
	{ "kst",   0,  32400, "Asia/Seoul"                    },
	{ "kst",   0,  30600, "Asia/Pyongyang"                },
	{ "kst",   0,  30600, "ROK"                           },
	{ "kst",   0,  32400, "ROK"                           },
	{ "lst",   1,   9394, "Europe/Riga"                   },
	{ "mddt",  1, -18000, "America/Cambridge_Bay"         },
	{ "mddt",  1, -18000, "America/Yellowknife"           },
	{ "mdst",  1,  16279, "Europe/Moscow"                 },
	{ "mdt",   1, -21600, "America/Denver"                },
	{ "mdt",   1, -21600, "America/Bahia_Banderas"        },
	{ "mdt",   1, -21600, "America/Boise"                 },
	{ "mdt",   1, -21600, "America/Cambridge_Bay"         },
	{ "mdt",   1, -21600, "America/Chihuahua"             },
	{ "mdt",   1, -21600, "America/Edmonton"              },
	{ "mdt",   1, -21600, "America/Hermosillo"            },
	{ "mdt",   1, -21600, "America/Inuvik"                },
	{ "mdt",   1, -21600, "America/Mazatlan"              },
	{ "mdt",   1, -21600, "America/North_Dakota/Beulah"   },
	{ "mdt",   1, -21600, "America/North_Dakota/Center"   },
	{ "mdt",   1, -21600, "America/North_Dakota/New_Salem" },
	{ "mdt",   1, -21600, "America/Ojinaga"               },
	{ "mdt",   1, -21600, "America/Phoenix"               },
	{ "mdt",   1, -21600, "America/Regina"                },
	{ "mdt",   1, -21600, "America/Shiprock"              },
	{ "mdt",   1, -21600, "America/Swift_Current"         },
	{ "mdt",   1, -21600, "America/Yellowknife"           },
	{ "mdt",   1, -21600, "Canada/Mountain"               },
	{ "mdt",   1, -21600, "Canada/Saskatchewan"           },
	{ "mdt",   1, -21600, "Mexico/BajaSur"                },
	{ "mest",  1,   7200, "MET"                           },
	{ "met",   0,   3600, "MET"                           },
	{ "mmt",   0,   9017, "Europe/Moscow"                 },
	{ "mmt",   0,   9079, "Europe/Moscow"                 },
	{ "mmt",   0, -13491, "America/Montevideo"            },
	{ "mmt",   0, -20712, "America/Managua"               },
	{ "mmt",   0,  -2588, "Africa/Monrovia"               },
	{ "mmt",   0,  -2670, "Africa/Monrovia"               },
	{ "mmt",   0,  17640, "Indian/Maldives"               },
	{ "mmt",   0,  19172, "Asia/Colombo"                  },
	{ "mmt",   0,  19270, "Asia/Calcutta"                 },
	{ "mmt",   0,  19270, "Asia/Kolkata"                  },
	{ "mmt",   0,  28656, "Asia/Makassar"                 },
	{ "mmt",   0,  28656, "Asia/Ujung_Pandang"            },
	{ "mmt",   0,   6600, "Europe/Minsk"                  },
	{ "mpt",   1, -21600, "America/Denver"                },
	{ "mpt",   1, -21600, "America/Boise"                 },
	{ "mpt",   1, -21600, "America/Cambridge_Bay"         },
	{ "mpt",   1, -21600, "America/Edmonton"              },
	{ "mpt",   1, -21600, "America/North_Dakota/Beulah"   },
	{ "mpt",   1, -21600, "America/North_Dakota/Center"   },
	{ "mpt",   1, -21600, "America/North_Dakota/New_Salem" },
	{ "mpt",   1, -21600, "America/Regina"                },
	{ "mpt",   1, -21600, "America/Shiprock"              },
	{ "mpt",   1, -21600, "America/Swift_Current"         },
	{ "mpt",   1, -21600, "America/Yellowknife"           },
	{ "mpt",   1, -21600, "Canada/Mountain"               },
	{ "mpt",   1, -21600, "Canada/Saskatchewan"           },
	{ "msd",   1,  14400, "Europe/Moscow"                 },
	{ "msd",   1,  14400, "Europe/Chisinau"               },
	{ "msd",   1,  14400, "Europe/Kaliningrad"            },
	{ "msd",   1,  14400, "Europe/Kiev"                   },
	{ "msd",   1,  14400, "Europe/Minsk"                  },
	{ "msd",   1,  14400, "Europe/Riga"                   },
	{ "msd",   1,  14400, "Europe/Simferopol"             },
	{ "msd",   1,  14400, "Europe/Tallinn"                },
	{ "msd",   1,  14400, "Europe/Tiraspol"               },
	{ "msd",   1,  14400, "Europe/Uzhgorod"               },
	{ "msd",   1,  14400, "Europe/Vilnius"                },
	{ "msd",   1,  14400, "Europe/Zaporozhye"             },
	{ "msk",   0,  10800, "Europe/Moscow"                 },
	{ "msk",   0,  14400, "Europe/Moscow"                 },
	{ "msk",   0,  10800, "Europe/Chisinau"               },
	{ "msk",   0,  10800, "Europe/Kaliningrad"            },
	{ "msk",   0,  10800, "Europe/Kiev"                   },
	{ "msk",   0,  10800, "Europe/Minsk"                  },
	{ "msk",   0,  10800, "Europe/Riga"                   },
	{ "msk",   0,  10800, "Europe/Simferopol"             },
	{ "msk",   0,  10800, "Europe/Tallinn"                },
	{ "msk",   0,  10800, "Europe/Tiraspol"               },
	{ "msk",   0,  10800, "Europe/Uzhgorod"               },
	{ "msk",   0,  10800, "Europe/Vilnius"                },
	{ "msk",   0,  10800, "Europe/Zaporozhye"             },
	{ "msk",   0,  14400, "Europe/Simferopol"             },
	{ "mst",   0, -25200, "America/Denver"                },
	{ "mst",   0, -25200, "America/Bahia_Banderas"        },
	{ "mst",   0, -25200, "America/Boise"                 },
	{ "mst",   0, -25200, "America/Cambridge_Bay"         },
	{ "mst",   0, -25200, "America/Chihuahua"             },
	{ "mst",   0, -25200, "America/Creston"               },
	{ "mst",   0, -25200, "America/Dawson_Creek"          },
	{ "mst",   0, -25200, "America/Edmonton"              },
	{ "mst",   0, -25200, "America/Ensenada"              },
	{ "mst",   0, -25200, "America/Fort_Nelson"           },
	{ "mst",   0, -25200, "America/Hermosillo"            },
	{ "mst",   0, -25200, "America/Inuvik"                },
	{ "mst",   0, -25200, "America/Mazatlan"              },
	{ "mst",   0, -25200, "America/Mexico_City"           },
	{ "mst",   0, -25200, "America/North_Dakota/Beulah"   },
	{ "mst",   0, -25200, "America/North_Dakota/Center"   },
	{ "mst",   0, -25200, "America/North_Dakota/New_Salem" },
	{ "mst",   0, -25200, "America/Ojinaga"               },
	{ "mst",   0, -25200, "America/Phoenix"               },
	{ "mst",   0, -25200, "America/Regina"                },
	{ "mst",   0, -25200, "America/Santa_Isabel"          },
	{ "mst",   0, -25200, "America/Shiprock"              },
	{ "mst",   0, -25200, "America/Swift_Current"         },
	{ "mst",   0, -25200, "America/Tijuana"               },
	{ "mst",   0, -25200, "America/Yellowknife"           },
	{ "mst",   0, -25200, "Canada/Mountain"               },
	{ "mst",   0, -25200, "Canada/Saskatchewan"           },
	{ "mst",   0, -25200, "Mexico/BajaNorte"              },
	{ "mst",   0, -25200, "Mexico/BajaSur"                },
	{ "mst",   0, -25200, "Mexico/General"                },
	{ "mst",   1,  12679, "Europe/Moscow"                 },
	{ "mwt",   1, -21600, "America/Denver"                },
	{ "mwt",   1, -21600, "America/Boise"                 },
	{ "mwt",   1, -21600, "America/Cambridge_Bay"         },
	{ "mwt",   1, -21600, "America/Edmonton"              },
	{ "mwt",   1, -21600, "America/North_Dakota/Beulah"   },
	{ "mwt",   1, -21600, "America/North_Dakota/Center"   },
	{ "mwt",   1, -21600, "America/North_Dakota/New_Salem" },
	{ "mwt",   1, -21600, "America/Phoenix"               },
	{ "mwt",   1, -21600, "America/Regina"                },
	{ "mwt",   1, -21600, "America/Shiprock"              },
	{ "mwt",   1, -21600, "America/Swift_Current"         },
	{ "mwt",   1, -21600, "America/Yellowknife"           },
	{ "mwt",   1, -21600, "Canada/Mountain"               },
	{ "mwt",   1, -21600, "Canada/Saskatchewan"           },
	{ "nddt",  1,  -5400, "America/St_Johns"              },
	{ "nddt",  1,  -5400, "Canada/Newfoundland"           },
	{ "ndt",   1,  -9052, "America/St_Johns"              },
	{ "ndt",   1,  -9000, "America/St_Johns"              },
	{ "ndt",   1,  -9000, "America/Goose_Bay"             },
	{ "ndt",   1,  -9000, "Canada/Newfoundland"           },
	{ "ndt",   1,  -9052, "America/Goose_Bay"             },
	{ "ndt",   1,  -9052, "Canada/Newfoundland"           },
	{ "npt",   1,  -9000, "America/St_Johns"              },
	{ "npt",   1, -36000, "America/Adak"                  },
	{ "npt",   1, -36000, "America/Atka"                  },
	{ "npt",   1, -36000, "America/Nome"                  },
	{ "npt",   1,  -9000, "America/Goose_Bay"             },
	{ "npt",   1,  -9000, "Canada/Newfoundland"           },
	{ "nst",   0, -12600, "America/St_Johns"              },
	{ "nst",   0, -12652, "America/St_Johns"              },
	{ "nst",   1,   4772, "Europe/Amsterdam"              },
	{ "nst",   0, -12600, "America/Goose_Bay"             },
	{ "nst",   0, -12600, "Canada/Newfoundland"           },
	{ "nst",   0, -12652, "America/Goose_Bay"             },
	{ "nst",   0, -12652, "Canada/Newfoundland"           },
	{ "nst",   0, -39600, "America/Adak"                  },
	{ "nst",   0, -39600, "America/Atka"                  },
	{ "nst",   0, -39600, "America/Nome"                  },
	{ "nwt",   1,  -9000, "America/St_Johns"              },
	{ "nwt",   1, -36000, "America/Adak"                  },
	{ "nwt",   1, -36000, "America/Atka"                  },
	{ "nwt",   1, -36000, "America/Nome"                  },
	{ "nwt",   1,  -9000, "America/Goose_Bay"             },
	{ "nwt",   1,  -9000, "Canada/Newfoundland"           },
	{ "nzdt",  1,  46800, "Pacific/Auckland"              },
	{ "nzdt",  1,  46800, "Antarctica/McMurdo"            },
	{ "nzdt",  1,  46800, "Antarctica/South_Pole"         },
	{ "nzdt",  1,  46800, "NZ"                            },
	{ "nzmt",  0,  41400, "Pacific/Auckland"              },
	{ "nzmt",  0,  41400, "Antarctica/McMurdo"            },
	{ "nzmt",  0,  41400, "Antarctica/South_Pole"         },
	{ "nzmt",  0,  41400, "NZ"                            },
	{ "nzst",  0,  43200, "Pacific/Auckland"              },
	{ "nzst",  1,  43200, "Pacific/Auckland"              },
	{ "nzst",  1,  45000, "Pacific/Auckland"              },
	{ "nzst",  0,  43200, "Antarctica/McMurdo"            },
	{ "nzst",  0,  43200, "Antarctica/South_Pole"         },
	{ "nzst",  0,  43200, "NZ"                            },
	{ "nzst",  1,  43200, "Antarctica/McMurdo"            },
	{ "nzst",  1,  43200, "Antarctica/South_Pole"         },
	{ "nzst",  1,  43200, "NZ"                            },
	{ "nzst",  1,  45000, "Antarctica/McMurdo"            },
	{ "nzst",  1,  45000, "Antarctica/South_Pole"         },
	{ "nzst",  1,  45000, "NZ"                            },
	{ "pddt",  1, -21600, "America/Inuvik"                },
	{ "pdt",   1, -25200, "America/Los_Angeles"           },
	{ "pdt",   1, -25200, "America/Boise"                 },
	{ "pdt",   1, -25200, "America/Dawson"                },
	{ "pdt",   1, -25200, "America/Dawson_Creek"          },
	{ "pdt",   1, -25200, "America/Ensenada"              },
	{ "pdt",   1, -25200, "America/Fort_Nelson"           },
	{ "pdt",   1, -25200, "America/Juneau"                },
	{ "pdt",   1, -25200, "America/Metlakatla"            },
	{ "pdt",   1, -25200, "America/Santa_Isabel"          },
	{ "pdt",   1, -25200, "America/Sitka"                 },
	{ "pdt",   1, -25200, "America/Tijuana"               },
	{ "pdt",   1, -25200, "America/Vancouver"             },
	{ "pdt",   1, -25200, "America/Whitehorse"            },
	{ "pdt",   1, -25200, "Canada/Pacific"                },
	{ "pdt",   1, -25200, "Canada/Yukon"                  },
	{ "pdt",   1, -25200, "Mexico/BajaNorte"              },
	{ "pdt",   1,  32400, "Asia/Manila"                   },
	{ "pkst",  1,  21600, "Asia/Karachi"                  },
	{ "pkt",   0,  18000, "Asia/Karachi"                  },
	{ "plmt",  0,  25590, "Asia/Ho_Chi_Minh"              },
	{ "plmt",  0,  25590, "Asia/Saigon"                   },
	{ "pmt",   0, -13236, "America/Paramaribo"            },
	{ "pmt",   0, -13252, "America/Paramaribo"            },
	{ "pmt",   0,  13505, "Asia/Yekaterinburg"            },
	{ "pmt",   0,  26240, "Asia/Pontianak"                },
	{ "pmt",   0,    561, "Africa/Algiers"                },
	{ "pmt",   0,    561, "Africa/Tunis"                  },
	{ "pmt",   0,    561, "Europe/Monaco"                 },
	{ "pmt",   0,    561, "Europe/Paris"                  },
	{ "ppmt",  0, -17340, "America/Port-au-Prince"        },
	{ "ppt",   1, -25200, "America/Los_Angeles"           },
	{ "ppt",   1, -25200, "America/Dawson_Creek"          },
	{ "ppt",   1, -25200, "America/Ensenada"              },
	{ "ppt",   1, -25200, "America/Fort_Nelson"           },
	{ "ppt",   1, -25200, "America/Juneau"                },
	{ "ppt",   1, -25200, "America/Metlakatla"            },
	{ "ppt",   1, -25200, "America/Santa_Isabel"          },
	{ "ppt",   1, -25200, "America/Sitka"                 },
	{ "ppt",   1, -25200, "America/Tijuana"               },
	{ "ppt",   1, -25200, "America/Vancouver"             },
	{ "ppt",   1, -25200, "Canada/Pacific"                },
	{ "ppt",   1, -25200, "Mexico/BajaNorte"              },
	{ "pst",   0, -28800, "America/Los_Angeles"           },
	{ "pst",   0, -28800, "America/Bahia_Banderas"        },
	{ "pst",   0, -28800, "America/Boise"                 },
	{ "pst",   0, -28800, "America/Creston"               },
	{ "pst",   0, -28800, "America/Dawson"                },
	{ "pst",   0, -28800, "America/Dawson_Creek"          },
	{ "pst",   0, -28800, "America/Ensenada"              },
	{ "pst",   0, -28800, "America/Fort_Nelson"           },
	{ "pst",   0, -28800, "America/Hermosillo"            },
	{ "pst",   0, -28800, "America/Inuvik"                },
	{ "pst",   0, -28800, "America/Juneau"                },
	{ "pst",   0, -28800, "America/Mazatlan"              },
	{ "pst",   0, -28800, "America/Metlakatla"            },
	{ "pst",   0, -28800, "America/Santa_Isabel"          },
	{ "pst",   0, -28800, "America/Sitka"                 },
	{ "pst",   0, -28800, "America/Tijuana"               },
	{ "pst",   0, -28800, "America/Vancouver"             },
	{ "pst",   0, -28800, "America/Whitehorse"            },
	{ "pst",   0, -28800, "Canada/Pacific"                },
	{ "pst",   0, -28800, "Canada/Yukon"                  },
	{ "pst",   0, -28800, "Mexico/BajaNorte"              },
	{ "pst",   0, -28800, "Mexico/BajaSur"                },
	{ "pst",   0,  28800, "Asia/Manila"                   },
	{ "pwt",   1, -25200, "America/Los_Angeles"           },
	{ "pwt",   1, -25200, "America/Dawson_Creek"          },
	{ "pwt",   1, -25200, "America/Ensenada"              },
	{ "pwt",   1, -25200, "America/Fort_Nelson"           },
	{ "pwt",   1, -25200, "America/Juneau"                },
	{ "pwt",   1, -25200, "America/Metlakatla"            },
	{ "pwt",   1, -25200, "America/Santa_Isabel"          },
	{ "pwt",   1, -25200, "America/Sitka"                 },
	{ "pwt",   1, -25200, "America/Tijuana"               },
	{ "pwt",   1, -25200, "America/Vancouver"             },
	{ "pwt",   1, -25200, "Canada/Pacific"                },
	{ "pwt",   1, -25200, "Mexico/BajaNorte"              },
	{ "qmt",   0, -18840, "America/Guayaquil"             },
	{ "rmt",   0,   5794, "Europe/Riga"                   },
	{ "rmt",   0,  23087, "Asia/Rangoon"                  },
	{ "rmt",   0,  23087, "Asia/Yangon"                   },
	{ "sast",  0,   7200, "Africa/Johannesburg"           },
	{ "sast",  1,  10800, "Africa/Johannesburg"           },
	{ "sast",  0,   5400, "Africa/Johannesburg"           },
	{ "sast",  1,  10800, "Africa/Maseru"                 },
	{ "sast",  1,  10800, "Africa/Mbabane"                },
	{ "sast",  1,  10800, "Africa/Windhoek"               },
	{ "sast",  0,   5400, "Africa/Maseru"                 },
	{ "sast",  0,   5400, "Africa/Mbabane"                },
	{ "sast",  0,   7200, "Africa/Maseru"                 },
	{ "sast",  0,   7200, "Africa/Mbabane"                },
	{ "sast",  0,   7200, "Africa/Windhoek"               },
	{ "sdmt",  0, -16800, "America/Santo_Domingo"         },
	{ "sjmt",  0, -20173, "America/Costa_Rica"            },
	{ "smt",   0, -13884, "Atlantic/Stanley"              },
	{ "smt",   0, -16966, "America/Punta_Arenas"          },
	{ "smt",   0, -16966, "America/Santiago"              },
	{ "smt",   0, -16966, "Chile/Continental"             },
	{ "smt",   0,  24925, "Asia/Kuala_Lumpur"             },
	{ "smt",   0,  24925, "Asia/Singapore"                },
	{ "smt",   0,   8160, "Europe/Simferopol"             },
	{ "sst",   0, -39600, "Pacific/Samoa"                 },
	{ "sst",   0, -39600, "Pacific/Midway"                },
	{ "sst",   0, -39600, "Pacific/Pago_Pago"             },
	{ "tbmt",  0,  10751, "Asia/Tbilisi"                  },
	{ "tmt",   0,  12344, "Asia/Tehran"                   },
	{ "tmt",   0,   5940, "Europe/Tallinn"                },
	{ "uct",   0,      0, "Etc/UCT"                       },
	{ "utc",   0,      0, "Etc/Universal"                 },
	{ "utc",   0,      0, "Etc/UTC"                       },
	{ "utc",   0,      0, "Etc/Zulu"                      },
	{ "utc",   0,      0, "UTC"                           },
	{ "utc",   0,      0, "UTC"                           },
	{ "wast",  1,   7200, "Africa/Windhoek"               },
	{ "wast",  1,   7200, "Africa/Ndjamena"               },
	{ "wat",   0,   3600, "Africa/Brazzaville"            },
	{ "wat",   0,   3600, "Africa/Bangui"                 },
	{ "wat",   0,   3600, "Africa/Douala"                 },
	{ "wat",   0,   3600, "Africa/Kinshasa"               },
	{ "wat",   0,   3600, "Africa/Lagos"                  },
	{ "wat",   0,   3600, "Africa/Libreville"             },
	{ "wat",   0,   3600, "Africa/Luanda"                 },
	{ "wat",   0,   3600, "Africa/Malabo"                 },
	{ "wat",   0,   3600, "Africa/Ndjamena"               },
	{ "wat",   0,   3600, "Africa/Niamey"                 },
	{ "wat",   0,   3600, "Africa/Porto-Novo"             },
	{ "wat",   0,   3600, "Africa/Sao_Tome"               },
	{ "wat",   1,   3600, "Africa/Windhoek"               },
	{ "wemt",  1,   7200, "Europe/Lisbon"                 },
	{ "wemt",  1,   7200, "Europe/Madrid"                 },
	{ "wemt",  1,   7200, "Europe/Monaco"                 },
	{ "wemt",  1,   7200, "Europe/Paris"                  },
	{ "west",  1,   3600, "Europe/Paris"                  },
	{ "west",  1,   7200, "Europe/Luxembourg"             },
	{ "west",  1,   3600, "Africa/Algiers"                },
	{ "west",  1,   3600, "Africa/Ceuta"                  },
	{ "west",  1,   3600, "Atlantic/Canary"               },
	{ "west",  1,   3600, "Atlantic/Faeroe"               },
	{ "west",  1,   3600, "Atlantic/Faroe"                },
	{ "west",  1,   3600, "Atlantic/Madeira"              },
	{ "west",  1,   3600, "Europe/Brussels"               },
	{ "west",  1,   3600, "Europe/Lisbon"                 },
	{ "west",  1,   3600, "Europe/Luxembourg"             },
	{ "west",  1,   3600, "Europe/Madrid"                 },
	{ "west",  1,   3600, "Europe/Monaco"                 },
	{ "wet",   0,      0, "Europe/Paris"                  },
	{ "wet",   0,   3600, "Europe/Luxembourg"             },
	{ "wet",   0,      0, "Africa/Algiers"                },
	{ "wet",   0,      0, "Africa/Ceuta"                  },
	{ "wet",   0,      0, "Atlantic/Azores"               },
	{ "wet",   0,      0, "Atlantic/Canary"               },
	{ "wet",   0,      0, "Atlantic/Faeroe"               },
	{ "wet",   0,      0, "Atlantic/Faroe"                },
	{ "wet",   0,      0, "Atlantic/Madeira"              },
	{ "wet",   0,      0, "Europe/Andorra"                },
	{ "wet",   0,      0, "Europe/Brussels"               },
	{ "wet",   0,      0, "Europe/Lisbon"                 },
	{ "wet",   0,      0, "Europe/Luxembourg"             },
	{ "wet",   0,      0, "Europe/Madrid"                 },
	{ "wet",   0,      0, "Europe/Monaco"                 },
	{ "wib",   0,  25200, "Asia/Jakarta"                  },
	{ "wib",   0,  25200, "Asia/Pontianak"                },
	{ "wita",  0,  28800, "Asia/Makassar"                 },
	{ "wita",  0,  28800, "Asia/Pontianak"                },
	{ "wita",  0,  28800, "Asia/Ujung_Pandang"            },
	{ "wit",   0,  32400, "Asia/Jayapura"                 },
	{ "wmt",   0,   5040, "Europe/Vilnius"                },
	{ "wmt",   0,   5040, "Europe/Warsaw"                 },
	{ "yddt",  1, -25200, "America/Dawson"                },
	{ "yddt",  1, -25200, "America/Whitehorse"            },
	{ "yddt",  1, -25200, "Canada/Yukon"                  },
	{ "ydt",   1, -28800, "America/Dawson"                },
	{ "ydt",   1, -28800, "America/Juneau"                },
	{ "ydt",   1, -28800, "America/Whitehorse"            },
	{ "ydt",   1, -28800, "America/Yakutat"               },
	{ "ydt",   1, -28800, "Canada/Yukon"                  },
	{ "ypt",   1, -28800, "America/Dawson"                },
	{ "ypt",   1, -28800, "America/Whitehorse"            },
	{ "ypt",   1, -28800, "America/Yakutat"               },
	{ "ypt",   1, -28800, "Canada/Yukon"                  },
	{ "yst",   0, -32400, "America/Anchorage"             },
	{ "yst",   0, -32400, "America/Dawson"                },
	{ "yst",   0, -32400, "America/Juneau"                },
	{ "yst",   0, -32400, "America/Nome"                  },
	{ "yst",   0, -32400, "America/Sitka"                 },
	{ "yst",   0, -32400, "America/Whitehorse"            },
	{ "yst",   0, -32400, "America/Yakutat"               },
	{ "yst",   0, -32400, "Canada/Yukon"                  },
	{ "ywt",   1, -28800, "America/Dawson"                },
	{ "ywt",   1, -28800, "America/Whitehorse"            },
	{ "ywt",   1, -28800, "America/Yakutat"               },
	{ "ywt",   1, -28800, "Canada/Yukon"                  },
	{ "a",     0,   3600, NULL                            },
	{ "b",     0,   7200, NULL                            },
	{ "c",     0,  10800, NULL                            },
	{ "d",     0,  14400, NULL                            },
	{ "e",     0,  18000, NULL                            },
	{ "f",     0,  21600, NULL                            },
	{ "g",     0,  25200, NULL                            },
	{ "h",     0,  28800, NULL                            },
	{ "i",     0,  32400, NULL                            },
	{ "k",     0,  36000, NULL                            },
	{ "l",     0,  39600, NULL                            },
	{ "m",     0,  43200, NULL                            },
	{ "n",     0,  -3600, NULL                            },
	{ "o",     0,  -7200, NULL                            },
	{ "p",     0, -10800, NULL                            },
	{ "q",     0, -14400, NULL                            },
	{ "r",     0, -18000, NULL                            },
	{ "s",     0, -21600, NULL                            },
	{ "t",     0, -25200, NULL                            },
	{ "u",     0, -28800, NULL                            },
	{ "v",     0, -32400, NULL                            },
	{ "w",     0, -36000, NULL                            },
	{ "x",     0, -39600, NULL                            },
	{ "y",     0, -43200, NULL                            },
	{ "z",     0,      0, NULL                            },
