EXPORTS
TidyLangPosixName
TidyLangWindowsName
getErrorCodeList
getInstalledLanguageList
getNextErrorCode
getNextInstalledLanguage
getNextStringKey
getNextWindowsLanguage
getStringKeyList
getWindowsLanguageList
tidyAccessWarningCount
tidyAttrDiscard
tidyAttrFirst
tidyAttrGetById
tidyAttrGetId
tidyAttrIsEvent
tidyAttrName
tidyAttrNext
tidyAttrValue
tidyBufAlloc
tidyBufAllocWithAllocator
tidyBufAppend
tidyBufAttach
tidyBufCheckAlloc
tidyBufClear
tidyBufDetach
tidyBufEndOfInput
tidyBufFree
tidyBufGetByte
tidyBufInit
tidyBufInitWithAllocator
tidyBufPopByte
tidyBufPutByte
tidyBufUngetByte
tidyCleanAndRepair
tidyConfigErrorCount
tidyCreate
tidyCreateWithAllocator
tidyDefaultString
tidyDetectedGenericXml
tidyDetectedHtmlVersion
tidyDetectedXhtml
tidyDiscardElement
tidyErrorCodeAsKey
tidyErrorCodeFromKey
tidyErrorCount
tidyErrorSummary
tidyFileExists
tidyGeneralInfo
tidyGetAppData
tidyGetArgFormat
tidyGetArgType
tidyGetArgValueDouble
tidyGetArgValueInt
tidyGetArgValueString
tidyGetArgValueUInt
tidyGetBody
tidyGetByte
tidyGetChild
tidyGetEmacsFile
tidyGetHead
tidyGetHtml
tidyGetLanguage
tidyGetMessage
tidyGetMessageArguments
tidyGetMessageCode
tidyGetMessageColumn
tidyGetMessageDefault
tidyGetMessageDoc
tidyGetMessageFormat
tidyGetMessageFormatDefault
tidyGetMessageIsMuted
tidyGetMessageKey
tidyGetMessageLevel
tidyGetMessageLine
tidyGetMessageOutput
tidyGetMessageOutputDefault
tidyGetMessagePos
tidyGetMessagePosDefault
tidyGetMessagePrefix
tidyGetMessagePrefixDefault
tidyGetNext
tidyGetNextMessageArgument
tidyGetNextOption
tidyGetOption
tidyGetOptionByName
tidyGetOptionList
tidyGetParent
tidyGetPrev
tidyGetRoot
tidyInitInputBuffer
tidyInitOutputBuffer
tidyInitSink
tidyInitSource
tidyIsEOF
tidyLibraryVersion
tidyLoadConfig
tidyLoadConfigEnc
tidyLocalizedString
tidyLocalizedStringN
tidyNodeColumn
tidyNodeGetId
tidyNodeGetName
tidyNodeGetText
tidyNodeGetType
tidyNodeGetValue
tidyNodeHasText
tidyNodeIsHeader
tidyNodeIsProp
tidyNodeIsText
tidyNodeLine
tidyOptCopyConfig
tidyOptDiffThanDefault
tidyOptDiffThanSnapshot
tidyOptGetBool
tidyOptGetCategory
tidyOptGetCurrPick
tidyOptGetDeclTagList
tidyOptGetDefault
tidyOptGetDefaultBool
tidyOptGetDefaultInt
tidyOptGetDoc
tidyOptGetDocLinksList
tidyOptGetEncName
tidyOptGetId
tidyOptGetIdForName
tidyOptGetInt
tidyOptGetMutedMessageList
tidyOptGetName
tidyOptGetNextDeclTag
tidyOptGetNextDocLinks
tidyOptGetNextMutedMessage
tidyOptGetNextPick
tidyOptGetNextPriorityAttr
tidyOptGetPickList
tidyOptGetPriorityAttrList
tidyOptGetType
tidyOptGetValue
tidyOptIsReadOnly
tidyOptParseValue
tidyOptResetAllToDefault
tidyOptResetToDefault
tidyOptResetToSnapshot
tidyOptSaveFile
tidyOptSaveSink
tidyOptSetBool
tidyOptSetInt
tidyOptSetValue
tidyOptSnapshot
tidyOptionIsList
tidyParseBuffer
tidyParseFile
tidyParseSource
tidyParseStdin
tidyParseString
tidyPlatform
tidyPutByte
tidyRelease
tidyReleaseDate
tidyReportDoctype
tidyRunDiagnostics
tidySaveBuffer
tidySaveFile
tidySaveSink
tidySaveStdout
tidySaveString
tidySetAppData
tidySetCharEncoding
tidySetConfigCallback
tidySetConfigChangeCallback
tidySetEmacsFile
tidySetErrorBuffer
tidySetErrorFile
tidySetErrorSink
tidySetFreeCall
tidySetInCharEncoding
tidySetLanguage
tidySetMallocCall
tidySetMessageCallback
tidySetOptionCallback
tidySetOutCharEncoding
tidySetPanicCall
tidySetPrettyPrinterCallback
tidySetReallocCall
tidySetReportCallback
tidySetReportFilter
tidyStatus
tidyUngetByte
tidyWarningCount
